# -----------------------------------------------------------------------------
# es7s/kolombos [Escape sequences and control characters visualiser]
# (C) 2022 A. Shavykin <0.delameter@gmail.com>
# -----------------------------------------------------------------------------
from __future__ import annotations

from typing import Deque, List, Tuple

from pytermor import Span, Spans, SequenceSGR
from pytermor.util import ReplaceSGR

from . import Chainable, SegmentPrinter, Segment, StartSequenceRef, StopSequenceRef, OneUseSequenceRef, SequenceRef
from .. import WaitRequest
from ...console import ConsoleDebugBuffer, Console
from ...settings import SettingsManager


# noinspection PyMethodMayBeStatic
class SegmentBuffer:
    def __init__(self):
        self._segment_chain: Deque[Chainable] = Deque[Chainable]()
        self._active_sgrs: List[SequenceSGR] = []
        self._last_detached_data_len = 0

        self._debug_buffer = ConsoleDebugBuffer('chainbuf')

    @property
    def data_len(self) -> int:
        return sum([el.data_len for el in self._segment_chain])

    @property
    def last_detached_data_len(self) -> int:
        return self._last_detached_data_len

    def attach(self, *segments: Segment):
        for segment in segments:
            f = Span(segment.opening_seq)
            if len(f.opening_seq.params) == 0 or set(f.opening_seq.params) == {0}:
                self._segment_chain.append(segment)
                continue

            self._segment_chain.extend([
                StartSequenceRef(f.opening_seq),
                segment,
                StopSequenceRef(f.opening_seq),
                OneUseSequenceRef(f.closing_seq)
            ])

    def detach_bytes(self, req_bytes: int, force: bool, printers: List[SegmentPrinter, ...]) -> Tuple[str, ...]:
        if self.data_len >= req_bytes or force:
            detached = self._detach(req_bytes)
            if len(detached) == 0:
                self._debug_buffer.write(1, 'Responsing with EOF')
                self._debug_buffer.write(2, f'Buffer state: {Console.printd(self)}')
                raise EOFError

            return self._format_multiple(detached, printers)

        self._debug_buffer.write(1, 'Responsing with WaitRequest')
        self._debug_buffer.write(2, f'Buffer state: {Console.printd(self)}')
        raise WaitRequest

    def detach_line(self, force: bool, printers: List[SegmentPrinter, ...]) -> Tuple[str, ...]:
        avail_bytes = 0
        has_newline = False
        for el in self._segment_chain:
            avail_bytes += el.data_len
            if el.is_newline:
                has_newline = True
                break

        if avail_bytes == 0:
            self._debug_buffer.write(1, 'Responsing with EOF')
            self._debug_buffer.write(2, f'Buffer state: {Console.printd(self)}')
            raise EOFError

        if has_newline or force:
            detached = self._detach(avail_bytes)
            return self._format_multiple(detached, printers)

        self._debug_buffer.write(1, 'Responsing with WaitRequest')
        self._debug_buffer.write(2, f'Buffer state: {Console.printd(self)}')
        raise WaitRequest

    def preview(self, max_input_len: int = 5) -> str:
        preview_data = self._preview_collect(max_input_len)
        raw_byte_len, sgr_byte_len, values, has_more = preview_data

        result = ('len ' +
                  Spans.BOLD(raw_byte_len) +
                  Spans.ITALIC(f'+{sgr_byte_len}'))

        if SettingsManager.app_settings.debug_buffer_contents:
            values_str = []
            has_more_str = '..' if has_more else ''
            for value in values:
                if isinstance(value, int):
                    values_str.append(f'{value:02x}')
                elif isinstance(value, SequenceRef):
                    values_str.append(Spans.ITALIC(self._preview_sgr(value.ref.assemble())))
            result += ': ' + Spans.GRAY('[' + ' '.join(values_str) + has_more_str + ']')

            if SettingsManager.app_settings.debug_buffer_contents_full:
                result += '. Active SGR buffer state: '
                sgrs_str = ' '.join([self._preview_sgr(sgr.assemble()) for sgr in self._active_sgrs])
                result += Spans.GRAY(f'[{sgrs_str}]')

        return result

    def _detach(self, req_bytes: int) -> List[Chainable]:
        self._debug_buffer.write(2, f'Buffer state: {Console.printd(self)}')

        if len(self._segment_chain) == 0:
            return []

        output = []
        output.extend([OneUseSequenceRef(sgr) for sgr in self._active_sgrs])
        while len(self._segment_chain) > 0:
            cur_element = self._segment_chain[0]

            if isinstance(cur_element, Segment):
                if req_bytes == 0:
                    break

                if cur_element.data_len <= req_bytes:
                    output.append(cur_element)
                    req_bytes -= cur_element.data_len
                    self._segment_chain.popleft()
                    continue

                else:
                    cur_element_left = cur_element.split(req_bytes)
                    output.append(cur_element_left)
                    break

            elif isinstance(cur_element, StartSequenceRef):
                if req_bytes == 0:
                    break
                output.append(cur_element)
                self._active_sgrs.append(cur_element.ref)

            elif isinstance(cur_element, StopSequenceRef):
                self._active_sgrs.remove(cur_element.ref)

            elif isinstance(cur_element, OneUseSequenceRef):
                output.append(cur_element)

            self._segment_chain.popleft()

        output.extend([OneUseSequenceRef(Span(sgr).closing_seq) for sgr in self._active_sgrs])
        self._last_detached_data_len = sum([el.data_len for el in output])

        self._debug_buffer.write(2, 'Detached ' + Spans.BOLD(sum([el.data_len for el in output])) + ' data byte(s)')
        self._debug_buffer.write(2, f'Buffer state: {Console.printd(self)}')
        return output

    def _format_multiple(self, detached: List[Chainable], printers: List[SegmentPrinter, ...]) -> Tuple[str, ...]:
        return tuple(self._format(detached, printer) for printer in printers)

    def _format(self, detached: List[Chainable], printer: SegmentPrinter) -> str:
        output = ''
        for cur_element in detached:
            if isinstance(cur_element, Segment):
                output += printer.print(cur_element)

            elif isinstance(cur_element, SequenceRef):  # StartSequenceRef | OneUseSequenceRef
                output += self._append_sgr_to_output(printer, cur_element.ref.assemble())

        # @TODO optimize sgrs: {RED}A{COLOR_OFF}{RED}A{COLOR_OFF}... -> {RED}AA{COLOR_OFF}
        return output

    def _append_sgr_to_output(self, printer: SegmentPrinter, sgr_str: str) -> str:
        if not printer.apply_sgr:
            return ''
        if printer.encode_sgr:
            return self._preview_sgr(sgr_str)
        return sgr_str

    def _get_active_sgrs_opening(self) -> str:
        return ''.join(sgr.assemble() for sgr in self._active_sgrs)

    def _get_active_sgrs_closing(self) -> str:
        return ''.join(Span(sgr).closing_str for sgr in self._active_sgrs)

    def _preview_sgr(self, sgr_str: str) -> str:
        return ReplaceSGR('[ǝ\\3]').apply(sgr_str)

    def _preview_collect(self, max_input_len: int = 5) -> Tuple[int, int, List[int | SequenceRef], bool]:
        raw_byte_len = 0
        sgr_byte_len = 0
        result: List[int | SequenceRef] = []
        max_input_exceeded = False
        for idx, el in enumerate(self._segment_chain):
            if idx >= max_input_len:
                max_input_exceeded = True

            if isinstance(el, Segment):
                if not max_input_exceeded:
                    result.extend(el.raw)
                raw_byte_len += el.data_len

            elif isinstance(el, SequenceRef):
                sgr_len = len(el.ref.assemble())
                if not max_input_exceeded:
                    result.append(el)
                sgr_byte_len += sgr_len

            else:
                raise RuntimeError(f'Unknown Chainable element: {el!r}')

        return raw_byte_len, sgr_byte_len, result, max_input_exceeded



# -----------------------------------------------------------------------------
# es7s/kolombos [Escape sequences and control characters visualiser]
# (C) 2022 A. Shavykin <0.delameter@gmail.com>
# -----------------------------------------------------------------------------
import abc

from .. import ParserBuffer, WHITESPACE_CHARCODES, PRINTABLE_CHARCODES
from ..segment import SegmentBuffer, SegmentPrinter, Segment


# noinspection PyMethodMayBeStatic
class AbstractFormatter(metaclass=abc.ABCMeta):
    def __init__(self, parser_buffer: ParserBuffer, segment_buffer: SegmentBuffer):
        self._parser_buffer = parser_buffer
        self._segment_buffer = segment_buffer

        self._raw_seg_printer = SegmentPrinter(True, False, self._seg_raw_to_hex)
        self._proc_seg_printer = SegmentPrinter(True, False, self._seg_processed_noop)
        self._origin_seg_printer = SegmentPrinter(False, False, self._seg_origin_noop)
        self._debug_raw_seg_printer = SegmentPrinter(False, False, self._seg_raw_to_hex)
        self._debug_proc_seg_printer = SegmentPrinter(False, False, self._seg_raw_to_safe)
        self._debug_sgr_seg_printer = SegmentPrinter(True, True, self._seg_raw_to_safe)

    @abc.abstractmethod
    def format(self): raise NotImplementedError

    def _seg_raw_to_hex(self, seg: Segment) -> str:
        return ''.join([f' {b:02x}' for b in seg.raw])

    def _seg_raw_to_safe(self, seg: Segment) -> str:
        result = ''
        for c in seg.raw.decode("ascii", errors="replace"):
            if ord(c) in WHITESPACE_CHARCODES:
                result += '·'
            elif ord(c) not in PRINTABLE_CHARCODES:
                result += '▯'
            else:
                result += c
        return result

    def _seg_processed_noop(self, seg: Segment) -> str:
        return seg.processed

    def _seg_origin_noop(self, seg: Segment) -> str:
        # @TODO shape into binary mode third column
        #   rename modes: text -> preview, binary -> breakdown
        #   breakdown columns: hex, chars, decoded
        return seg.raw.decode('utf-8', errors='replace').replace('\n', '.')



# -----------------------------------------------------------------------------
# es7s/kolombos [Escape sequences and control characters visualiser]
# (C) 2022 A. Shavykin <0.delameter@gmail.com>
# -----------------------------------------------------------------------------
import re
from argparse import HelpFormatter, Action, ArgumentParser, SUPPRESS
from typing import Optional, Iterable, List

from pytermor import Spans

from .byteio import Reader
from .byteio.template import Template


class CustomHelpFormatter(HelpFormatter):
    INDENT_INCREMENT = 2
    INDENT = ' ' * INDENT_INCREMENT

    @staticmethod
    def format_header(title: str) -> str:
        return Spans.BOLD(title.upper())

    def __init__(self, prog):
        super().__init__(prog, max_help_position=30, indent_increment=self.INDENT_INCREMENT)

    def start_section(self, heading: Optional[str]):
        super().start_section(self.format_header(heading))

    def add_usage(self, usage: Optional[str], actions: Iterable[Action], groups: Iterable,
                  prefix: Optional[str] = ...):
        super().add_text(self.format_header('usage'))

        usage = usage.replace("\n", f"\n{self.INDENT}")
        super().add_usage(usage, actions, groups, prefix=self.INDENT)

    def add_examples(self, examples: List[str]):
        self.start_section('example' + ('s' if len(examples) > 1 else ''))
        self._add_item(self._format_text, ['\n'.join(examples)])
        self.end_section()

    def _format_action_invocation(self, action):
        # same as in superclass, but without printing argument for short options
        if not action.option_strings:
            default = self._get_default_metavar_for_positional(action)
            metavar, = self._metavar_formatter(action, default)(1)
            return metavar
        else:
            parts = []
            if action.nargs == 0:
                parts.extend(action.option_strings)
            else:
                default = self._get_default_metavar_for_optional(action)
                args_string = self._format_args(action, default)
                for option_string in action.option_strings:
                    if len(option_string) > 2 or len(action.option_strings) == 1:
                        parts.append(f'{option_string} {args_string}')
                    else:
                        parts.append(option_string)

            return ', '.join(parts)

    def _format_text(self, text: str) -> str:
        return super()._format_text(text).rstrip('\n') + '\n'

    def _fill_text(self, text, width, indent):
        return ''.join(indent + line for line in text.splitlines(keepends=True))


class CustomArgumentParser(ArgumentParser):
    def __init__(self, examples: List[str] = None, epilog: List[str] = None, usage: List[str] = None, **kwargs):
        self.examples = examples
        kwargs.update({
            'epilog': '\n'.join(epilog),
            'usage': '\n'.join(usage),
        })
        super(CustomArgumentParser, self).__init__(**kwargs)

    def format_help(self) -> str:
        formatter = self._get_formatter()
        if self.epilog:
            formatter.add_text(' ')
            formatter.add_text(self.epilog)
        if self.examples and isinstance(formatter, CustomHelpFormatter):
            formatter.add_examples(self.examples)

        ending_formatted = formatter.format_help()
        self.epilog = None

        result = super().format_help() + ending_formatted
        # remove ':' from headers ('<_b>header:<_f>'):
        result = re.sub(r'(\033\[[0-9;]*m)?\s*:\s*(\n|\033|$)', r'\1\2', result)
        return result


class AppArgumentParser(CustomArgumentParser):
    def __init__(self):
        fmt_b = Spans.BOLD
        fmt_u = Spans.UNDERLINED
        fmt_default = Spans.YELLOW

        super().__init__(
            description='Escape sequences and control characters visualiser',
            usage=[
                '%(prog)s [[--text] | --binary] [<options>] [<file>]',
                '%(prog)s --legend',
                '%(prog)s --version',
                '%(prog)s --help',
            ],
            epilog=[
                'Mandatory or optional arguments to long options are also mandatory or optional for any'
                ' corresponding short options. Arguments can be separated with both space or "=" in both cases.',
                '',
                f'Binary mode disables {fmt_b("--marker")} setting, because marker length should be always equal to actual sequence length. Therefore, control chars have 0 details level, while escape seqs are displayed with full details (2). Also, debug mode sets {fmt_b("--buffer")} setting to {Reader.READ_CHUNK_SIZE_DEBUG} bytes (however, it can be overriden as usual).',
                '',
                '(c) 2022 A. Shavykin <0.delameter@gmail.com>',
            ],
            examples=[
                'Read file in text mode, highlight escape sequences, marker verbosity 2 (max)',
                ''.ljust(4) + f"{fmt_u('%(prog)s')} --focus-esc -m{fmt_u(2)} file.txt",
                '',
                'Read file in binary mode, ignore whitespace, printables and UTF-8 sequences',
                ''.ljust(4) + f"{fmt_u('%(prog)s')} --binary -USP file.bin",
                '',
                'Read stdin in binary mode, highlight control characters, process 128 first bytes and exit',
                ''.ljust(4) + f"{fmt_u('%(prog)s')} --binary --focus-control -B{fmt_u(128)}",
                '',
                'Display annotation symbol list and color map',
                ''.ljust(4) + f"{fmt_u('%(prog)s')} --legend",
                '\n'
            ],
            add_help=False,
            formatter_class=lambda prog: CustomHelpFormatter(prog),
            prog='kolombos'
        )

        self.add_argument('filename', metavar='<file>', nargs='?', help='file to read from; if empty or "-", read stdin instead')

        modes_group = self.add_argument_group('operating mode')
        modes_group_nested = modes_group.add_mutually_exclusive_group()
        # modes_group_nested.add_argument('-a', '--auto', action='store_true', default=True, help='open file in text mode, fallback to binary on failure (default)')
        modes_group_nested.add_argument('-t', '--text', action='store_true', default=True, help='open file in text mode '+fmt_default('[this is a default]'))
        modes_group_nested.add_argument('-b', '--binary', action='store_true', default=False, help='open file in binary mode')
        modes_group_nested.add_argument('-l', '--legend', action='store_true', default=False, help='show annotation symbol list and exit')
        modes_group.add_argument('-v', '--version', action='store_true', default=False, help='show app version and exit')
        modes_group.add_argument('-h', '--help', action='help', default=SUPPRESS, help='show this help message and exit')

        char_class_group = self.add_argument_group('character class options')
        space_output_group = char_class_group.add_mutually_exclusive_group()
        control_output_group = char_class_group.add_mutually_exclusive_group()
        printable_output_group = char_class_group.add_mutually_exclusive_group()
        esc_output_group = char_class_group.add_mutually_exclusive_group()
        utf8_output_group = char_class_group.add_mutually_exclusive_group()
        binary_output_group = char_class_group.add_mutually_exclusive_group()
        space_output_group.add_argument('-s', '--focus-space', action='store_true', default=False, help='highlight whitespace markers')
        control_output_group.add_argument('-c', '--focus-control', action='store_true', default=False, help='highlight control char markers')
        printable_output_group.add_argument('-p', '--focus-printable', action='store_true', default=False, help='highlight printable chars')
        esc_output_group.add_argument('-e', '--focus-esc', action='store_true', default=False, help='highlight escape sequences markers')
        utf8_output_group.add_argument('-u', '--focus-utf8', action='store_true', default=False, help='highlight UTF-8 sequences')
        binary_output_group.add_argument('-i', '--focus-binary', action='store_true', default=False, help='highlight binary data bytes')
        space_output_group.add_argument('-S', '--ignore-space', action='store_true', default=False, help='dim/hide whitespaces')
        control_output_group.add_argument('-C', '--ignore-control', action='store_true', default=False, help='dim/hide control chars')
        printable_output_group.add_argument('-P', '--ignore-printable', action='store_true', default=False, help='dim/hide printable chars')
        esc_output_group.add_argument('-E', '--ignore-esc', action='store_true', default=False, help='dim/hide escape sequences')
        utf8_output_group.add_argument('-U', '--ignore-utf8', action='store_true', default=False, help='dim/hide UTF-8 sequences')
        binary_output_group.add_argument('-I', '--ignore-binary', action='store_true', default=False, help='dim/hide binary data')

        generic_group = self.add_argument_group('generic options')
        generic_group.add_argument('-f', '--buffer', metavar='<size>', type=int, default=None, help='read buffer size, in bytes '+fmt_default(f'[default: {Reader.READ_CHUNK_SIZE}]'))
        generic_group.add_argument('-L', '--max-lines', metavar='<num>', action='store', type=int, default=0, help='stop after reading <num> lines '+fmt_default('[default: no limit]'))
        generic_group.add_argument('-B', '--max-bytes', metavar='<num>', action='store', type=int, default=0, help='stop after reading <num> bytes '+fmt_default('[default: no limit]'))
        generic_group.add_argument('-D', '--debug', action='count', default=0, help='enable debug mode; can be used from 1 to 4 times, each level increases verbosity (-D|DD|DDD|DDDD)')
        generic_group.add_argument('--color-markers', action='store_true', default=False, help='apply SGR marker format to themselves')

        text_mode_group = self.add_argument_group('text mode options')
        text_mode_group.add_argument('-m', '--marker', metavar='<details>', action='store', type=int, default=0, help='marker details: 0 is none, 1 is brief, 2 is full '+fmt_default('[default: %(default)s]'))
        text_mode_group.add_argument('--no-separators', action='store_true', default=False, help='do not print '+Template.wrap_in_separators('separators')+' around escape sequences')
        # text_mode_group.add_argument('-Q', '--squash-ignored', action='store_true', default=False, help=Span(Seqs.HI_YELLOW)('TODO ')+'replace sequences of ignored characters with one character')
        # text_mode_group.add_argument('-H', '--hide-ignored', action='store_true', default=False, help=Span(Seqs.HI_YELLOW)('TODO ')+'completely hide ignored character classes')
        text_mode_group.add_argument('--no-line-numbers', action='store_true', default=False, help='do not print line numbers')

        bin_mode_group = self.add_argument_group('binary mode options')
        bin_mode_group.add_argument('-w', '--columns', metavar='<num>', action='store', type=int, default=0, help='format output as <num>-columns wide table '+fmt_default('[default: auto]'))
        bin_mode_group.add_argument('-d', '--decode', action='store_true', default=False, help='decode valid UTF-8 sequences, print as unicode chars')
        offsets_group = bin_mode_group.add_mutually_exclusive_group()
        offsets_group.add_argument('--decimal-offsets', action='store_true', default=False, help='output offsets in decimal format '+fmt_default('[default: hex format]'))
        offsets_group.add_argument('--no-offsets', action='store_true', default=False, help='do not print offsets')



# -----------------------------------------------------------------------------
#  pytermor [ANSI formatted terminal output toolset]
#  (c) 2022. A. Shavykin <0.delameter@gmail.com>
# -----------------------------------------------------------------------------
"""
Some of the Python Standard Library methods rewritten
for correct work with strings containing control sequences.
"""
from .string_filter import ReplaceSGR


def ljust_sgr(s: str, width: int, fillchar: str = ' ') -> str:
    """
    SGR-formatting-aware implementation of ``str.ljust``.

    Return a left-justified string of length ``width``. Padding is done
    using the specified fill character (default is a space).
    """
    sanitized = ReplaceSGR().apply(s)
    return s + fillchar * max(0, width - len(sanitized))


def rjust_sgr(s: str, width: int, fillchar: str = ' ') -> str:
    """
    SGR-formatting-aware implementation of ``str.rjust``.

    Return a right-justified string of length ``width``. Padding is done
    using the specified fill character (default is a space).
    """
    sanitized = ReplaceSGR().apply(s)
    return fillchar * max(0, width - len(sanitized)) + s


def center_sgr(s: str, width: int, fillchar: str = ' ') -> str:
    """
    SGR-formatting-aware implementation of ``str.center``.

    Return a centered string of length ``width``. Padding is done using the
    specified fill character (default is a space).

    .. todo ::

        поверить корректность работы в случае эмодзи (напр. 🔋)
        если алгоритм поедет -- можно заменить на f-стринги
    """
    sanitized = ReplaceSGR().apply(s)
    fill_len = max(0, width - len(sanitized))
    if fill_len == 0:
        return s

    right_fill_len = fill_len // 2
    left_fill_len = fill_len - right_fill_len
    return (fillchar * left_fill_len) + s + (fillchar * right_fill_len)



# -----------------------------------------------------------------------------
#  pytermor [ANSI formatted terminal output toolset]
#  (c) 2022. A. Shavykin <0.delameter@gmail.com>
# -----------------------------------------------------------------------------
"""
String filtering module.

Main idea is to provide a common interface for string filtering, that can make
possible working with filters like with objects rather than with functions/lambdas.

.. testsetup:: *

    from pytermor.ansi import Spans
    from pytermor.util.string_filter import apply_filters, ReplaceSGR, VisualuzeWhitespace

"""
from __future__ import annotations

import re
from functools import reduce
from re import Match, Pattern
from typing import Generic, AnyStr, Type, Callable

from ..ansi import Spans


SGR_REGEXP = re.compile(r'(\x1b)(\[)(([0-9;])*)(m)')


def apply_filters(s: AnyStr, *args: StringFilter[AnyStr]|Type[StringFilter[AnyStr]]) -> AnyStr:  # @FIXME StringFilter[AnyStr] -> StringFilter ?
    """
    Method for applying dynamic filter list to a target string/bytes.
    Example (will replace all :kbd:`ESC` control characters to :kbd:`E` and
    thus make SGR params visible):

    >>> apply_filters(Spans.RED('test'), ReplaceSGR(r'E\\2\\3\\5'))
    'E[31mtestE[39m'

    Note that type of ``s`` argument must be same as ``StringFilter`` parameterized
    type, i.e. :class:`ReplaceNonAsciiBytes` is ``StringFilter[bytes]`` type, so
    you can apply it only to bytes-type strings.

    :param AnyStr s: String to filter.
    :param args:     `StringFilter` instance(s) or ``StringFilter`` class(es).
    :return:         Filtered ``s``.
    """
    filters = map(lambda t: t() if isinstance(t, type) else t, args)
    return reduce(lambda s_, f: f.apply(s_), filters, s)


class StringFilter(Generic[AnyStr]):
    """
    Common string modifier interface.
    """
    def __init__(
        self,
        pattern: AnyStr|Pattern[AnyStr],
        repl: AnyStr|Callable[[AnyStr|Match], AnyStr]
    ):
        if isinstance(pattern, (str, bytes)):
            self._regex = re.compile(pattern)
        else:
            self._regex = pattern
        self._repl = repl

    def __call__(self, s: AnyStr) -> AnyStr:
        """ Can be used instead of `apply()` """
        return self.apply(s)

    def apply(self, s: AnyStr) -> AnyStr:
        """ Apply filter to ``s`` string (or bytes). """
        return self._regex.sub(self._repl, s)


class VisualuzeWhitespace(StringFilter[str]):
    """
    Replace every invisible character with ``repl`` (default is :kbd:`·`),
    except newlines. Newlines are kept and get prepneded with same string.

    >>> VisualuzeWhitespace().apply('A  B  C')
    'A··B··C'
    >>> apply_filters('1. D\\n2. L ', VisualuzeWhitespace)
    '1.·D·\\n2.·L·'
    """
    def __init__(self, repl: AnyStr = '·'):
        super().__init__(r'(\n)|\s', repl + '\\1')


class ReplaceSGR(StringFilter[str]):
    """
    Find all SGR seqs (e.g. :kbd:`ESC[1;4m`) and replace with given string. More
    specific version of :class:`ReplaceCSI`.

    :param repl:
        Replacement, can contain regexp groups (see :meth:`apply_filters()`).
    """
    def __init__(self, repl: AnyStr = ''):
        super().__init__(SGR_REGEXP, repl)


class ReplaceCSI(StringFilter[str]):
    """
    Find all CSI seqs (i.e. :kbd:`ESC[*`) and replace with given string. Less
    specific version of :class:`ReplaceSGR`, as CSI consists of SGR and many other
    sequence subtypes.

    :param repl:
        Replacement, can contain regexp groups (see :meth:`apply_filters()`).
    """
    def __init__(self, repl: AnyStr = ''):
        super().__init__(r'(\x1b)(\[)(([0-9;:<=>?])*)([@A-Za-z])', repl)


class ReplaceNonAsciiBytes(StringFilter[bytes]):
    """
    Keep 7-bit ASCII bytes [:kbd:`0x00` - :kbd:`0x7f`], replace other to :kbd:`?`.

    :param repl: Replacement byte-string.
    """
    def __init__(self, repl: AnyStr = b'?'):
        super().__init__(b'[\x80-\xff]', repl)



# -----------------------------------------------------------------------------
# es7s/kolombos [Escape sequences and control characters visualiser]
# (C) 2022 A. Shavykin <0.delameter@gmail.com>
# -----------------------------------------------------------------------------
from importlib.resources import read_text

from . import AbstractRunner
from ..console import Console


class LegendRunner(AbstractRunner):
    LEGEND_FILENAME = 'legend.ansi'

    def run(self):
        for line in read_text('kolombos', self.LEGEND_FILENAME).splitlines(keepends=True):
            if line.startswith('#'):
                continue
            Console.info(line, end='')



# -----------------------------------------------------------------------------
# es7s/kolombos [Escape sequences and control characters visualiser]
# (C) 2022 A. Shavykin <0.delameter@gmail.com>
# -----------------------------------------------------------------------------
import re
from os.path import join, dirname, realpath, abspath

import pytermor
from . import AbstractRunner
from ..version import __version__
from ..console import Console


class VersionRunner(AbstractRunner):
    def run(self):
        Console.info("es7s/kolombos".ljust(16) + __version__)
        Console.info("pytermor".ljust(16) + pytermor.__version__)



# -----------------------------------------------------------------------------
# es7s/kolombos [Escape sequences and control characters visualiser]
# (C) 2022 A. Shavykin <0.delameter@gmail.com>
# -----------------------------------------------------------------------------
from .chainable import Chainable
from .segment import Segment
from .printer import SegmentPrinter
from .sequence_ref import SequenceRef, StartSequenceRef, StopSequenceRef, OneUseSequenceRef
from .buffer import SegmentBuffer



# -----------------------------------------------------------------------------
# es7s/kolombos [Escape sequences and control characters visualiser]
# (C) 2022 A. Shavykin <0.delameter@gmail.com>
# -----------------------------------------------------------------------------
from __future__ import annotations

from typing import List

from pytermor import SequenceSGR, Seqs, NOOP_SEQ

from . import Template
from .. import CharClass, MarkerDetailsEnum, OpeningSeqPOV, LabelPOV, ReadMode
from ..segment import Segment
from ...settings import SettingsManager


class EscapeSequenceTemplate(Template):
    ESQ_MARKER_LABEL_SEQ: SequenceSGR = NOOP_SEQ

    def __init__(self, opening_seq: SequenceSGR | OpeningSeqPOV, label: str | LabelPOV = ''):
        super().__init__(CharClass.ESCAPE_SEQ, opening_seq, label)

    def substitute(self, raw: bytes) -> List[Segment]:
        if self._display_mode.is_ignored:
            return super().substitute(raw)

        self._substituted.clear()

        label_raw = raw[:1]   # label of escape sequence is always 1-byte
        params_raw = raw[1:]  # @TODO replace space in CSI params to placeholder

        primary_seg = self._create_primary_segment(
            self._get_label_opening_seq(),
            label_raw,
            self._process(label_raw)
        )
        self._substituted.insert(0, primary_seg)
        self._fill_additional_segments(params_raw)

        if (self._read_mode != ReadMode.BINARY and
            not SettingsManager.app_settings.no_separators and (
                self._marker_details is MarkerDetailsEnum.BRIEF_DETAILS or
                self._marker_details is MarkerDetailsEnum.FULL_DETAILS
            )):
            self._wrap_in_separators()

        return self._substituted

    def _fill_additional_segments(self, params_raw: bytes):
        if self._marker_details is MarkerDetailsEnum.NO_DETAILS:
            return

        params_processed = params_raw.decode('ascii', errors='ignore')

        if self._marker_details is MarkerDetailsEnum.BRIEF_DETAILS:
            params_processed = self._get_brief_details_processed()

        self._substituted.append(
            self._create_details_segment(params_raw, params_processed)
        )

    def _get_brief_details_processed(self) -> str:
        return ''

    def _get_label_opening_seq(self) -> SequenceSGR:
        return self._opening_seq_stack.get(self._display_mode, self._read_mode) + self.ESQ_MARKER_LABEL_SEQ

    def _get_details_opening_seq(self) -> SequenceSGR:
        return self._opening_seq_stack.get() + self.DETAILS_OPENING_SEQ

    def _wrap_in_separators(self):
        self.wrap_in_separators(self._substituted)



# -----------------------------------------------------------------------------
# es7s/kolombos [Escape sequences and control characters visualiser]
# (C) 2022 A. Shavykin <0.delameter@gmail.com>
# -----------------------------------------------------------------------------
from pytermor import Seqs
from pytermor.util import ReplaceSGR

from . import AbstractFormatter
from .. import ParserBuffer, WaitRequest
from ..segment import SegmentBuffer
from ...console import ConsoleDebugBuffer, ConsoleOutputBuffer


class TextFormatter(AbstractFormatter):
    def __init__(self, parser_buffer: ParserBuffer, segment_buffer: SegmentBuffer):
        super().__init__(parser_buffer, segment_buffer)

        self._offset = 0
        self._line_num = 1

        self._output_buffer = ConsoleOutputBuffer()
        self._debug_buffer = ConsoleDebugBuffer('txtfmt', Seqs.YELLOW)

    def format(self):
        while True:
            try:
                self._debug_buffer.write(1, 'Requested line')
                force = self._parser_buffer.closed
                result = self._segment_buffer.detach_line(force, [
                    self._debug_sgr_seg_printer,
                    self._debug_raw_seg_printer,
                    self._debug_proc_seg_printer,
                    self._proc_seg_printer,
                ])
            except WaitRequest:
                break
            except EOFError:
                break

            debug_sgr_line, debug_raw_line, debug_proc_line, final_proc_line = result

            self._debug_buffer.write(3, debug_sgr_line, offset=self._offset)
            self._debug_buffer.write(2, debug_raw_line, offset=self._offset)
            self._debug_buffer.write(1, debug_proc_line, offset=self._offset)
            self._output_buffer.write_with_line_num(final_proc_line, line_num=self._line_num)

            if not ReplaceSGR().apply(final_proc_line).endswith('\n'):
                self._output_buffer.write('', end='\n')

            self._offset += self._segment_buffer.last_detached_data_len
            self._line_num += 1

        if self._parser_buffer.closed:
            self._debug_buffer.write(1, 'EOF')



# -----------------------------------------------------------------------------
# es7s/kolombos [Escape sequences and control characters visualiser]
# (C) 2022 A. Shavykin <0.delameter@gmail.com>
# -----------------------------------------------------------------------------
from enum import Enum
from typing import Dict, Union

CONTROL_CHARCODES = list(range(0x00, 0x09)) + list(range(0x0e, 0x20)) + list(range(0x7f, 0x100))
WHITESPACE_CHARCODES = list(range(0x09, 0x0e)) + [0x20]
PRINTABLE_CHARCODES = list(range(0x21, 0x7f))
BINARY_CHARCODES = list(range(0x80, 0x100))
CHARCODE_TO_SAFE_CHAR_MAP = {   # @TODO впилить в отладчоный вывод AbstractFormatter._seg_raw_to_safe
    **{b: chr(b) for b in range(0x00, 0x100)},
    **{b: '·' for b in WHITESPACE_CHARCODES},
    **{b: '▯' for b in CONTROL_CHARCODES},
    **{b: '▯' for b in BINARY_CHARCODES},
}


class ReadMode(Enum):
    BINARY = 'binary'
    TEXT = 'text'

    @property
    def is_binary(self) -> bool: return self is self.BINARY
    @property
    def is_text(self) -> bool: return self is self.TEXT


class DisplayMode(Enum):
    DEFAULT = 'default'
    FOCUSED = 'focused'
    IGNORED = 'ignored'

    @property
    def is_default(self) -> bool: return self is self.DEFAULT
    @property
    def is_focused(self) -> bool: return self is self.FOCUSED
    @property
    def is_ignored(self) -> bool: return self is self.IGNORED


ByteIOMode = Union[DisplayMode, ReadMode]


class CharClass(Enum):
    CONTROL_CHAR = 'control'
    ESCAPE_SEQ = 'esc'
    WHITESPACE = 'space'
    UTF_8_SEQ = 'utf8'
    BINARY_DATA = 'binary'
    PRINTABLE_CHAR = 'printable'


TYPE_LABEL_MAP: Dict[CharClass, str] = {
    CharClass.CONTROL_CHAR: 'C',
    CharClass.ESCAPE_SEQ: 'E',
    CharClass.WHITESPACE: 'S',
    CharClass.UTF_8_SEQ: 'U',
    CharClass.BINARY_DATA: 'B',
    CharClass.PRINTABLE_CHAR: 'P',
}
TYPE_LABEL_DETAILS = '*'


class MarkerDetailsEnum(Enum):
    NO_DETAILS = 0
    BRIEF_DETAILS = 1
    FULL_DETAILS = 2
    BINARY_STRICT = 3  # require len(raw) = len(processed)
    DEFAULT = NO_DETAILS



# -----------------------------------------------------------------------------
# es7s/kolombos [Escape sequences and control characters visualiser]
# (C) 2022 A. Shavykin <0.delameter@gmail.com>
# -----------------------------------------------------------------------------
from ._abstract import AbstractRunner

from .byteio import ByteIoRunner
from .legend import LegendRunner
from .version import VersionRunner

from .factory import RunnerFactory



# -----------------------------------------------------------------------------
#  pytermor [ANSI formatted terminal output toolset]
#  (c) 2022. A. Shavykin <0.delameter@gmail.com>
# -----------------------------------------------------------------------------
from __future__ import annotations

import os
import sys
from typing import TypeVar, Generic, AnyStr

T = TypeVar('T')
""" Any """


class Registry(Generic[T]):
    """
    Registry of elements of specified type.
    """
    @classmethod
    def resolve(cls, name: str) -> T:
        """
        Case-insensitive search through registry contents.

        :param name: name of the value to look up for.
        :return:     value or KeyError if nothing found.
        """
        name_norm = name.upper()
        if (value := getattr(cls, name_norm, None)) is not None:
            return value
        raise KeyError(f'No item named "{name_norm}" (<- "{name}") is found '
                       f'in {cls.__name__} registry')


def get_terminal_width() -> int:
    """
    get_terminal_width
    :return:  terminal_width
    """
    try:
        import shutil as _shutil
        return _shutil.get_terminal_size().columns - 2
    except ImportError:
        return 80


def wait_key() -> AnyStr|None:
    """
    Wait for a key press on the console and return it.
    """
    if os.name == 'nt':
        import msvcrt
        return msvcrt.getch()

    import termios
    fd = sys.stdin.fileno()

    oldterm = termios.tcgetattr(fd)
    newattr = termios.tcgetattr(fd)
    newattr[3] = newattr[3] & ~termios.ICANON & ~termios.ECHO
    termios.tcsetattr(fd, termios.TCSANOW, newattr)

    result = None
    try:
        result = sys.stdin.read(1)
    except IOError:
        pass
    finally:
        termios.tcsetattr(fd, termios.TCSAFLUSH, oldterm)
    return result


class LogicError(Exception):
    pass



# -----------------------------------------------------------------------------
#  pytermor [ANSI formatted terminal output toolset]
#  (c) 2022. A. Shavykin <0.delameter@gmail.com>
# -----------------------------------------------------------------------------
from .ansi import IntCodes, Seqs, Spans, SequenceSGR, Span, NOOP_SEQ, NOOP_SPAN
from .color import Colors, ColorIndexed16, ColorIndexed256, ColorRGB, NOOP_COLOR
from .render import Styles, Style, Text, SgrRenderer, RendererManager, NOOP_STYLE
from ._version import __version__

import logging
from logging import NullHandler

logging.getLogger(__name__).addHandler(NullHandler())



# -----------------------------------------------------------------------------
# es7s/kolombos [Escape sequences and control characters visualiser]
# (C) 2022 A. Shavykin <0.delameter@gmail.com>
# -----------------------------------------------------------------------------
from __future__ import annotations

from typing import Dict, List

from pytermor import SequenceSGR

from . import Template
from .. import CharClass, MarkerDetailsEnum, OpeningSeqPOV, LabelPOV
from ..segment import Segment


class ControlCharGenericTemplate(Template):
    def __init__(self, opening_seq: SequenceSGR | OpeningSeqPOV, label: str | LabelPOV = ''):
        super().__init__(CharClass.CONTROL_CHAR, opening_seq, label)

        self._brief_label_map: Dict[int, str] = {
            0x01: 'A', 0x02: 'B', 0x03: 'C',  0x04: 'D', 0x05: 'E', 0x06: 'F',
            0x07: 'G', 0x0e: 'N', 0x0f: 'O',  0x10: 'P', 0x11: 'Q', 0x12: 'R',
            0x13: 'S', 0x14: 'T', 0x15: 'U',  0x16: 'V', 0x17: 'W', 0x18: 'X',
            0x19: 'Y', 0x1a: 'Z', 0x1c: '\\', 0x1d: ']', 0x1e: '^', 0x1f: '_',
        }

    def substitute(self, raw: bytes) -> List[Segment]:
        self._substituted.clear()

        if self._marker_details is MarkerDetailsEnum.BINARY_STRICT or \
           self._marker_details is MarkerDetailsEnum.NO_DETAILS:
            self._substituted.append(self._create_primary_segment(
                self._opening_seq_stack.get(self._display_mode, self._read_mode),
                bytes(raw),
                self._process(raw),
            ))
            return self._substituted

        for label_raw in raw:
            label_processed = self._process_byte(label_raw)
            self._substituted.append(self._create_primary_segment(
                self._opening_seq_stack.get(self._display_mode, self._read_mode),
                bytes(label_raw),
                label_processed,
            ))
            self._fill_additional_segments(label_raw)

        return self._substituted

    def _fill_additional_segments(self, b: int):
        if self._display_mode.is_ignored:
            return

        if self._marker_details is MarkerDetailsEnum.BINARY_STRICT or \
           self._marker_details is MarkerDetailsEnum.NO_DETAILS:
            return

        elif self._marker_details is MarkerDetailsEnum.BRIEF_DETAILS:
            if b not in self._brief_label_map.keys():
                raise KeyError(f'No brief label defined for byte 0x{b:02x}')
            self._substituted.append(self._create_details_segment(b'', self._brief_label_map[b]))

        elif self._marker_details is MarkerDetailsEnum.FULL_DETAILS:
            self._substituted.append(self._create_details_segment(b'', '{:02x}'.format(b)))

        else:
            raise RuntimeError(f'Invalid marker details level: {self._marker_details.value}')

    def _get_details_opening_seq(self) -> SequenceSGR:
        return self._opening_seq_stack.get() + self.DETAILS_OPENING_SEQ



# -----------------------------------------------------------------------------
# es7s/kolombos [Escape sequences and control characters visualiser]
# (C) 2022 A. Shavykin <0.delameter@gmail.com>
# -----------------------------------------------------------------------------
from . import AbstractRunner
from ..byteio import ParserBuffer, Parser, Reader
from ..byteio.formatter import FormatterFactory
from ..byteio.segment import SegmentBuffer
from ..byteio.template import TemplateRegistry
from ..console import Console, ConsoleDebugBuffer
from ..settings import SettingsManager


# noinspection PyMethodMayBeStatic
class ByteIoRunner(AbstractRunner):
    def run(self):
        # try:
        self._reinit()
        self._reader.read()

        # except BinaryDataError:
        #    self._reader.close()
        #    self._switch_mode_or_exit()

        #    settings.binary = True ; settings.text = False
        #    self._reinit()
        #    self._reader.read()

    def _reinit(self):
        self._debug_buffer = ConsoleDebugBuffer()
        if not SettingsManager.app_settings.debug_settings:
            self._debug_buffer.write(1, Console.get_separator_line(main_open=True))

        self._parser_buffer = ParserBuffer()
        self._segment_buffer = SegmentBuffer()
        self._template_registry = TemplateRegistry()

        self._reader = Reader(SettingsManager.app_settings.filename, self._process_chunk_buffered)
        self._parser = Parser(self._parser_buffer, self._segment_buffer, self._template_registry)
        self._formatter = FormatterFactory.create(self._parser_buffer, self._segment_buffer)

    def _process_chunk_buffered(self, raw_input: bytes, offset: int, finish: bool):
        self._parser_buffer.append_raw(raw_input, finish)
        self._parser.parse(offset)
        self._formatter.format()

        self._debug_buffer.write(1, Console.get_separator_line(main_close=finish))


    # def _switch_mode_or_exit(self):
    #     Console.warn('Binary data detected, cannot proceed in text mode')
    #
    #     if not SettingsManager.app_settings.auto:
    #         Console.info('Use -b option to run in binary mode')
    #         App.exit(2)
    #
    #     if self._reader.reading_stdin:
    #         Console.info('Cannot repeat input stream reading because it\'s stdin')
    #         App.exit(3)



# -----------------------------------------------------------------------------
# es7s/kolombos [Escape sequences and control characters visualiser]
# (C) 2022 A. Shavykin <0.delameter@gmail.com>
# -----------------------------------------------------------------------------
from __future__ import annotations

import abc
import re
import sys
from typing import Callable, IO

from pytermor import Spans, Seqs

from ..console import ConsoleDebugBuffer, Console
from ..settings import SettingsManager


class Reader(metaclass=abc.ABCMeta):
    READ_CHUNK_SIZE: int = 4096
    READ_CHUNK_SIZE_DEBUG: int = 128

    def __init__(self, filename: str|None, read_callback: Callable[[bytes, int, bool], None]):
        self._filename = filename
        self._io: IO|None = None
        self._offset = 0
        self._chunk_size = self._get_chunk_size()
        self._read_callback = read_callback
        self._debug_buffer = ConsoleDebugBuffer('reader', Seqs.MAGENTA)

    @property
    def reading_stdin(self) -> bool:
        return not self._filename or self._filename == '-'

    def read(self):
        self._open()
        self._debug_buffer.write(2, f'Read buffer: size {Spans.BOLD(self._chunk_size)}')

        max_bytes: int|None = SettingsManager.app_settings.max_bytes
        max_lines: int|None = SettingsManager.app_settings.max_lines
        lines = 0
        chunks = 0
        try:
            while raw_input := self._io.read(self._chunk_size):
                self._debug_buffer.write(1, f'Read chunk #{chunks}: {Console.printd(raw_input, 5)}', offset=self._offset)
                self._debug_buffer.write(3, f'Current position: {Spans.BOLD(self._io.tell() if not self.reading_stdin else "n/a")}', offset=self._offset)
                chunks += 1
                if max_lines:
                    for line_break in re.finditer(b'\x0a', raw_input):
                        lines += 1
                        if lines >= max_lines:
                            raw_input = raw_input[:line_break.end()-1]
                            self._debug_buffer.write(2, 'Line limit exceeded: ' + Spans.BOLD(max_lines), offset=self._offset)
                            self._debug_buffer.write(3, f'Cropping input -> {Console.printd(raw_input)}', offset=self._offset)
                            break

                if max_bytes and self._offset + len(raw_input) > max_bytes:
                    raw_input = raw_input[:max_bytes - self._offset]
                    self._debug_buffer.write(2, 'Byte limit exceeded: ' + Spans.BOLD(max_bytes), offset=self._offset)
                    self._debug_buffer.write(3, f'Cropping input -> {Console.printd(raw_input)}', offset=self._offset)
                    break

                self._read_callback(raw_input, self._offset, False)
                self._offset += len(raw_input)
                raw_input = b''

                if (max_bytes and self._offset >= max_bytes) or \
                   (max_lines and lines >= max_lines):
                    break

            self._debug_buffer.write(1, f'Encountered EOF', offset=self._offset)
            self._read_callback(raw_input, self._offset, True)

        except KeyboardInterrupt:
            pass
        finally:
            self.close()

    def _open(self):
        if self.reading_stdin:
            self._io = sys.stdin.buffer
            self._debug_buffer.write(1, f'Reading from stdin')
        else:
            self._io = open(self._filename, 'rb')
            self._debug_buffer.write(1, f'Opened file: {Spans.BOLD(self._filename)}')

    def _get_chunk_size(self) -> int:
        if SettingsManager.app_settings.buffer:
            return int(SettingsManager.app_settings.buffer)
        if SettingsManager.app_settings.debug > 0:
            return self.READ_CHUNK_SIZE_DEBUG
        return self.READ_CHUNK_SIZE

    def close(self):
        if self._io and not self._io.closed:
            self._io.close()



# -----------------------------------------------------------------------------
# es7s/kolombos [Escape sequences and control characters visualiser]
# (C) 2022 A. Shavykin <0.delameter@gmail.com>
# -----------------------------------------------------------------------------
from .common import WaitRequest
from .const import ReadMode, DisplayMode, ByteIOMode, CharClass, MarkerDetailsEnum, CONTROL_CHARCODES, \
    BINARY_CHARCODES, PRINTABLE_CHARCODES, WHITESPACE_CHARCODES, CHARCODE_TO_SAFE_CHAR_MAP
from .partial_override import PartialOverride, OpeningSeqPOV, LabelPOV

from .reader import Reader
from .parser import ParserBuffer, Parser



# -----------------------------------------------------------------------------
# es7s/kolombos [Escape sequences and control characters visualiser]
# (C) 2022 A. Shavykin <0.delameter@gmail.com>
# -----------------------------------------------------------------------------
from __future__ import annotations

from . import AppArgumentParser
from .console import Console
from .runner import RunnerFactory
from .settings import SettingsManager


# noinspection PyMethodMayBeStatic
class App:
    def run(self):
        try:
            self._parse_args()  # help processing is handled by argparse
            (RunnerFactory.create()).run()
        except Exception as e:
            Console.on_exception(e)
            self._exit(1)
        self._exit(0)

    def _parse_args(self):
        SettingsManager.init()
        AppArgumentParser().parse_args(namespace=SettingsManager.app_settings)
        Console.debug_settings()

    def _exit(self, code: int):
        print()
        exit(code)



# -----------------------------------------------------------------------------
# es7s/kolombos [Escape sequences and control characters visualiser]
# (C) 2022 A. Shavykin <0.delameter@gmail.com>
# -----------------------------------------------------------------------------
from __future__ import annotations

from pytermor import SequenceSGR

from . import Chainable


class Segment(Chainable):
    def __init__(self, opening_seq: SequenceSGR, type_label: str, raw: bytes, processed: str = None):
        self._opening_seq = opening_seq
        self._type_label = type_label
        self._raw = raw
        self._processed = processed

    def split(self, num_bytes: int) -> Segment:
        if not self.is_consistent:
            raise BufferError('Unknown how to split inconsistent segment. Usually this error means that app is operating'
                              'in binary mode, but for some reason template processing resulted in output with '
                              'different length (which is allowed for text mode only).')

        left = Segment(self._opening_seq, self._type_label, self._raw[:num_bytes], self._processed[:num_bytes])

        self._raw = self._raw[num_bytes:]
        self._processed = self._processed[num_bytes:]
        return left

    @property
    def data_len(self) -> int:
        return len(self._raw)

    @property
    def is_newline(self) -> bool:
        return '\n' in self._processed

    @property
    def is_consistent(self) -> bool:
        return len(self._raw) == len(self._processed)

    @property
    def type_label(self) -> str:
        return self._type_label

    @property
    def opening_seq(self) -> SequenceSGR:
        return self._opening_seq

    @property
    def raw(self) -> bytes:
        return self._raw

    @property
    def processed(self) -> str:
        return self._processed

    def __eq__(self, other: Segment):
        if not isinstance(other, Segment):
            return False

        return self._opening_seq == other._opening_seq \
               and self._type_label == other._type_label \
               and self._raw == other._raw \
               and self._processed == other._processed

    def __repr__(self):
        return f'{self.__class__.__name__}[{self._raw.hex(" ")}]->[{self._processed}]'



# -----------------------------------------------------------------------------
# es7s/kolombos [Escape sequences and control characters visualiser]
# (C) 2022 A. Shavykin <0.delameter@gmail.com>
# -----------------------------------------------------------------------------

from .template import Template

from .template_control import ControlCharGenericTemplate
from .template_printable import PrintableCharTemplate
from .template_utf8 import Utf8SequenceTemplate
from .template_newline import NewlineTemplate
from .template_escape import EscapeSequenceTemplate
from .template_escape_sgr import EscapeSequenceSGRTemplate

from .registry import TemplateRegistry



# -----------------------------------------------------------------------------
# es7s/kolombos [Escape sequences and control characters visualiser]
# (C) 2022 A. Shavykin <0.delameter@gmail.com>
# -----------------------------------------------------------------------------
from __future__ import annotations

from typing import Callable

from . import Segment


class SegmentPrinter:
    def __init__(self, apply_sgr: bool, encode_sgr: bool, format_fn: Callable[[Segment], str]):
        self._apply_sgr = apply_sgr
        self._encode_sgr = encode_sgr
        self._format_fn = format_fn

    @property
    def apply_sgr(self):
        return self._apply_sgr

    @property
    def encode_sgr(self):
        return self._encode_sgr

    def print(self, segment: Segment) -> str:
        return self._format_fn(segment)



# -----------------------------------------------------------------------------
# es7s/kolombos [Escape sequences and control characters visualiser]
# (C) 2022 A. Shavykin <0.delameter@gmail.com>
# -----------------------------------------------------------------------------
from __future__ import annotations

from typing import List

from pytermor import SequenceSGR, Seqs, ansi

from .. import CharClass, DisplayMode, MarkerDetailsEnum, ReadMode, PartialOverride, OpeningSeqPOV, LabelPOV
from ..const import TYPE_LABEL_MAP, TYPE_LABEL_DETAILS
from ..segment import Segment
from ...settings import SettingsManager


class Template:
    SEPARATOR_LEFT =  '⢸'
    SEPARATOR_RIGHT = '⡇'

    IGNORED_LABEL: str = '×'
    IGNORED_OPENING_SEQ: SequenceSGR = Seqs.GRAY + Seqs.DIM
    DETAILS_OPENING_SEQ: SequenceSGR = SequenceSGR.init_color_indexed(16, True) + Seqs.UNDERLINED

    def __init__(self, char_class: CharClass, opening_seq: SequenceSGR | OpeningSeqPOV, label: str | LabelPOV = ''):
        if not isinstance(opening_seq, PartialOverride):
            opening_seq = OpeningSeqPOV(opening_seq)
        if not isinstance(label, PartialOverride):
            label = LabelPOV(label)

        self._char_class: CharClass = char_class
        self._opening_seq_stack: OpeningSeqPOV = opening_seq
        self._label_stack: LabelPOV = label
        self._substituted: List[Segment] = []

        self._display_mode: DisplayMode = DisplayMode.DEFAULT
        self._read_mode: ReadMode = ReadMode.TEXT
        self._marker_details: MarkerDetailsEnum = MarkerDetailsEnum.NO_DETAILS
        self._decode: bool = False

        if not self._opening_seq_stack.has_key(DisplayMode.FOCUSED):
            self._opening_seq_stack.set(DisplayMode.FOCUSED, self._opening_seq_stack.get() + Seqs.INVERSED)
        if not self._opening_seq_stack.has_key(DisplayMode.IGNORED):
            self._opening_seq_stack.set(DisplayMode.IGNORED, self.IGNORED_OPENING_SEQ)

        if not self._label_stack.has_key(DisplayMode.IGNORED):
            self._label_stack.set(DisplayMode.IGNORED, self.IGNORED_LABEL)

        self.update_settings()

    def update_settings(self):
        app_settings = SettingsManager.app_settings
        self._display_mode = app_settings.get_char_class_display_mode(self._char_class)
        self._read_mode = app_settings.read_mode
        self._marker_details = app_settings.effective_marker_details
        self._decode = app_settings.decode

    def substitute(self, raw: bytes) -> List[Segment]:
        self._substituted.clear()

        primary_seg = self._create_primary_segment(
            self._opening_seq_stack.get(self._display_mode, self._read_mode),
            raw,
            self._process(raw)  # <-- self._substituted can be changed here
        )
        self._substituted.insert(0, primary_seg)
        return self._substituted

    @property
    def char_class(self) -> CharClass:
        return self._char_class

    @property
    def label_stack(self) -> LabelPOV:
        return self._label_stack

    def _process(self, raw: bytes) -> str:
        return ''.join(self._process_byte(b) for b in raw)

    def _process_byte(self, b: int) -> str:
        return self._label_stack.get(self._display_mode, self._read_mode)

    def _create_primary_segment(self, opening_seq: SequenceSGR, raw: bytes, processed: str) -> Segment:
        return Segment(opening_seq, self._get_type_label(), raw, processed)

    def _create_details_segment(self, raw: bytes, processed: str) -> Segment:
        return Segment(self._get_details_opening_seq(), TYPE_LABEL_DETAILS, raw, processed)

    def _get_type_label(self) -> str:
        return TYPE_LABEL_MAP[self._char_class]

    def _get_details_opening_seq(self) -> SequenceSGR:
        raise NotImplemented

    @staticmethod
    def wrap_in_separators(s: str|List[Segment]) -> str|None:
        if isinstance(s, str):
            return f'{Template.SEPARATOR_LEFT}{s}{Template.SEPARATOR_RIGHT}'

        if isinstance(s, list):
            s.insert(0, Segment(SequenceSGR.init_color_indexed(255), '', b'', Template.SEPARATOR_LEFT))
            s.append(Segment(SequenceSGR.init_color_indexed(255), '', b'', Template.SEPARATOR_RIGHT))
            return

        raise TypeError(f'Invalid argument type {type(s)}')



# -----------------------------------------------------------------------------
# es7s/kolombos [Escape sequences and control characters visualiser]
# (C) 2022 A. Shavykin <0.delameter@gmail.com>
# -----------------------------------------------------------------------------
from __future__ import annotations

from pytermor import SequenceSGR, IntCodes, Seqs

from . import EscapeSequenceTemplate
from .. import OpeningSeqPOV, LabelPOV, MarkerDetailsEnum
from ...settings import SettingsManager


# noinspection PyMethodMayBeStatic
class EscapeSequenceSGRTemplate(EscapeSequenceTemplate):
    M1_SEPARATOR = '･'

    ALLOWED_SGRS_FOR_MARKER_FORMAT = [  # INVERSED, BOLD, UNDERLINED are reserved for markers themself
        IntCodes.DIM, IntCodes.ITALIC,
        IntCodes.OVERLINED, IntCodes.CROSSLINED,
        *IntCodes.LIST_ALL_COLORS,
    ]

    def __init__(self, opening_seq: SequenceSGR | OpeningSeqPOV, label: str | LabelPOV = ''):
        super().__init__(opening_seq, label)
        self._details_fmt_str: str|None = None
        self._brief_details_processed: str|None = None
        self._details_opening_seq: SequenceSGR|None = None

    def get_details_fmt_str(self) -> str:
        if self._details_fmt_str is None:
            raise ValueError('Details format not initialzied')
        return self._details_fmt_str

    def set_details_fmt_str(self, s: bytes):
        self._details_fmt_str = s.decode('ascii')

        params = self._details_fmt_str.split(';')
        pcodes = [int(p) for p in params if p]
        pcodes_allowed = []
        brief_names = []
        while len(pcodes):  # @TODO automatically set black/white text if only bg color is defined, and vice versa
            pcode = pcodes.pop(0)

            if pcode == IntCodes.COLOR_EXTENDED or pcode == IntCodes.BG_COLOR_EXTENDED:
                brief_type = 'F' if pcode == IntCodes.COLOR_EXTENDED else 'B'
                if len(pcodes) >= 2 and pcodes[0] == IntCodes.EXTENDED_MODE_256:
                    pcodes_allowed.extend([pcode, pcodes.pop(0), pcodes.pop(0)])
                    brief_names.append(f'{brief_type}{pcodes_allowed[-1]}')
                    continue
                if len(pcodes) >= 4 and pcodes[0] == IntCodes.EXTENDED_MODE_RGB:
                    pcodes_allowed.extend([pcode, *[pcodes.pop(0) for _ in range(4)]])
                    brief_names.append(
                        f'{brief_type}'
                        f'{pcodes_allowed[-3] * 256 * 256 + pcodes_allowed[-2] * 256 + pcodes_allowed[-1]:06X}'
                    )
                    continue
            elif pcode in self.ALLOWED_SGRS_FOR_MARKER_FORMAT:
                pcodes_allowed.append(pcode)
            brief_names.append(pcode)

        self._brief_details_processed = f'{Seqs.UNDERLINED_OFF}{self.M1_SEPARATOR}{Seqs.UNDERLINED}'.join(str(bn) for bn in brief_names)
        self._details_opening_seq = SequenceSGR(*pcodes_allowed)

    def _get_brief_details_processed(self) -> str:
        if self._brief_details_processed is not None:
            return self._brief_details_processed
        return self.get_details_fmt_str()

    def _get_label_opening_seq(self) -> SequenceSGR:
        if SettingsManager.app_settings.color_markers:
            return self._details_opening_seq + self.ESQ_MARKER_LABEL_SEQ
        return super()._get_label_opening_seq()

    def _get_details_opening_seq(self) -> SequenceSGR:
        if not SettingsManager.app_settings.color_markers:
            return self.DETAILS_OPENING_SEQ
        return self.DETAILS_OPENING_SEQ + self._details_opening_seq



# -----------------------------------------------------------------------------
# es7s/kolombos [Escape sequences and control characters visualiser]
# (C) 2022 A. Shavykin <0.delameter@gmail.com>
# -----------------------------------------------------------------------------
from abc import ABCMeta, abstractmethod


class Chainable(metaclass=ABCMeta):
    @property
    @abstractmethod
    def data_len(self) -> int: raise NotImplementedError

    @property
    @abstractmethod
    def is_newline(self) -> bool: raise NotImplementedError



# -----------------------------------------------------------------------------
# es7s/kolombos [Escape sequences and control characters visualiser]
# (C) 2022 A. Shavykin <0.delameter@gmail.com>
# -----------------------------------------------------------------------------
from __future__ import annotations

from pytermor import Spans

from ...console import ConsoleDebugBuffer, Console


class ParserBuffer:
    def __init__(self):
        self._raw_buffer: bytes = b''
        self._closed: bool = False

        self._debug_buffer = ConsoleDebugBuffer('parsbuf')

    def append_raw(self, b: bytes, finish: bool = False):
        self._raw_buffer += b
        self._closed = finish

        self._debug_buffer.write(3, f'Appending {Spans.BOLD(len(b))} bytes')
        self._debug_buffer.write(2, f'Buffer state: {Console.printd(self._raw_buffer)}')
        if finish:
            self._debug_buffer.write(1, 'Closing buffer for input')

    def get_raw(self) -> bytes:
        return self._raw_buffer

    def crop_raw(self, new_buffer: bytes):
        if len(new_buffer) == 0:
            self._raw_buffer = b''
            self._debug_buffer.write(3, f'Purging')
            self._debug_buffer.write(2, f'Buffer state: {Console.printd(self._raw_buffer)}')
            return

        offset_delta = self._raw_buffer.find(new_buffer)
        if offset_delta == -1:
            raise RuntimeError(f'New buffer is not a part of the current one: {Console.printd(new_buffer, 32)}')

        self._raw_buffer = self._raw_buffer[offset_delta:]
        self._debug_buffer.write(3, f'Cropping')
        self._debug_buffer.write(2, f'Buffer state: {Console.printd(self._raw_buffer)}')

    @property
    def closed(self) -> bool: return self._closed



# -----------------------------------------------------------------------------
#  pytermor [ANSI formatted terminal output toolset]
#  (c) 2022. A. Shavykin <0.delameter@gmail.com>
# -----------------------------------------------------------------------------
"""
Module with output formatters. By default :class:`SgrRenderer` is used. It
also contains compatibility settings, see `SgrRenderer.set_up()`.

Working with non-default renderer can be achieved in two ways:

a. Method `RendererManager.set_up()` sets the default renderer globally. After that
   calling ``str(<Renderable>)`` will automatically invoke said renderer and all
   formatting will be applied.
b. Alternatively, you can use renderer's own class method `IRenderer.render()`
   directly and avoid messing up with the manager: ``HtmlRenderer.render(<Renderable>)``

.. rubric:: TL;DR

To unconditionally print formatted message to output terminal, do something like this:

>>> from pytermor import SgrRenderer, Styles, Text
>>> SgrRenderer.set_up(force_styles=True)
>>> print(Text('Warning: AAAA', Styles.WARNING))
\x1b[33mWarning: AAAA\x1b[39m


.. testsetup:: *

    from pytermor.color import Colors
    from pytermor.render import *

    SgrRenderer.set_up(force_styles=True)

-----

.. todo ::

    Scheme can be simplified, too many unnecessary abstractions for now.

    Renderable
          (implemented by Text) should include algorithms for creating intermediate
          styles for text pieces that lie in beetween first opening sequence (or tag)
          and second, for example -- this happens when one Text instance is included
          into another.
    Style's
          responsibility is to preserve the state of text piece and thats it.
    Renderer
          should transform style into corresponding output format and thats it.

    API 2:
        Text(string, style, leave_open = True)  # no style means all open styles will
        be closed at the end
        Text().append(string, style, leave_open = True)
        Text().prepend(string, style, leave_open = False)
        Text().raw
        Text().apply(style)
        Text().render(with=IRenderer())
        Text() + Text() = Text().append(Text().raw, Text().style)
        Text() + str = Text().append(str)
        str + Text() = Text().prepend(str)

        Style(style, fg, bg...)
        # no Style().render()!
        IRenderer().setup()
        IRenderer().render(text)
        SgrRenderer().is_sgr_usage_allowed()

    renderers should have instance methods only!
"""
from __future__ import annotations

import sys
from abc import abstractmethod, ABCMeta
from dataclasses import dataclass, field
from functools import reduce
from typing import List, Sized, Any
from typing import Type, Dict, Set

from pytermor.common import LogicError
from .ansi import SequenceSGR, Span, NOOP_SEQ, Seqs, IntCodes
from .color import Color, ColorRGB, ColorIndexed16, ColorIndexed256, NOOP_COLOR, Colors
from .common import Registry
from .util.string_filter import ReplaceSGR


class Renderable(Sized, metaclass=ABCMeta):
    """
    Renderable abstract class. Can be inherited if the default style
    overlaps resolution mechanism implemented in `Text` is not good enough
    and you want to implement your own.
    """
    def render(self) -> str:
        return self._render_using(RendererManager.get_default())

    @abstractmethod
    def _render_using(self, renderer: Type[IRenderer]) -> str:
        raise NotImplemented

    def __str__(self) -> str:
        return self.render()

    @abstractmethod
    def __len__(self) -> int:
        raise NotImplemented


class Text(Renderable):
    """
    Text
    """

    def __init__(self, text: Any = None, style: Style|str = None):
        self._runs = [self._TextRun(text, style)]

    def _render_using(self, renderer: Type[IRenderer]) -> str:
        return ''.join(run._render_using(renderer) for run in self._runs)

    def raw(self) -> str:
        return ''.join(run.raw() for run in self._runs)

    def append(self, text: str|Text):
        if isinstance(text, str):
            self._runs.append(self._TextRun(text))
        elif isinstance(text, Text):
            self._runs += text._runs
        else:
            raise TypeError('Can add Text to another Text instance or str only')

    def prepend(self, text: str|Text):
        if isinstance(text, str):
            self._runs.insert(0, self._TextRun(text))
        elif isinstance(text, Text):
            self._runs = text._runs + self._runs
        else:
            raise TypeError('Can add Text to another Text instance or str only')

    def __len__(self) -> int:
        return sum(len(r) for r in self._runs)

    def __format__(self, *args, **kwargs) -> str:
        runs_amount = len(self._runs)
        if runs_amount == 0:
            return ''.__format__(*args, **kwargs)
        if runs_amount > 1:
            raise RuntimeError(
                f'Can only __format__ Texts consisting of 0 or 1 TextRuns, '
                f'got {runs_amount}. Consider applying the styles and creating'
                f' the Text instance after value formatting.')
        return self._runs[0].__format__(*args, **kwargs)

    def __add__(self, other: str|Text) -> Text:
        self.append(other)
        return self

    def __iadd__(self, other: str|Text) -> Text:
        self.append(other)
        return self

    def __radd__(self, other: str|Text) -> Text:
        self.prepend(other)
        return self

    class _TextRun(Renderable):
        def __init__(self, string: Any = None, style: Style|str = None):
            self._string: str = str(string) if string else ''
            if isinstance(style, str):
                style = Style(fg=style)
            self._style: Style|None = style

        def _render_using(self, renderer: Type[IRenderer]) -> str:
            if not self._style:
                return self._string
            return renderer.render(self._string, self._style)

        def raw(self) -> str:
            return self._string

        def __len__(self) -> int:
            return len(self._string)

        def __format__(self, *args, **kwargs) -> str:
            self._string = self._string.__format__(*args, **kwargs)
            return self.render()


@dataclass
class Style:
    """Create a new ``Style()``.

    Key difference between ``Styles`` and ``Spans`` or ``SGRs`` is that
    ``Styles`` describe colors in RGB format and therefore support output
    rendering in several different formats (see :mod:`._render`).

    Both ``fg`` and ``bg`` can be specified as:

    1. :class:`.Color` instance or library preset;
    2. key code -- name of any of aforementioned presets, case-insensitive;
    3. integer color value in hexademical RGB format.
    4. None -- the color will be unset.

    :param fg:          Foreground (i.e., text) color.
    :param bg:          Background color.
    :param inherit:     Parent instance to copy unset properties from.
    :param blink:       Blinking effect; *supported by limited amount of Renderers*.
    :param bold:        Bold or increased intensity.
    :param crosslined:  Strikethrough.
    :param dim:         Faint, decreased intensity.
    :param double_underlined:
        Faint, decreased intensity.
    :param inversed:    Swap foreground and background colors.
    :param italic:      Italic.
    :param overlined:   Overline.
    :param underlined:  Underline.
    :param class_name:  Arbitary string used by some renderers, e.g. by ``HtmlRenderer``.

    >>> Style(fg='green', bold=True)
    Style[fg=008000, no bg, bold]
    >>> Style(bg=0x0000ff)
    Style[no fg, bg=0000ff]
    >>> Style(fg=Colors.DEEP_SKY_BLUE_1, bg=Colors.GREY_93)
    Style[fg=00afff, bg=eeeeee]
    """
    _fg: Color = field(default=None, init=False)
    _bg: Color = field(default=None, init=False)

    def __init__(self, inherit: Style = None, fg: Color|int|str = None,
                 bg: Color|int|str = None, blink: bool = None, bold: bool = None,
                 crosslined: bool = None, dim: bool = None,
                 double_underlined: bool = None, inversed: bool = None,
                 italic: bool = None, overlined: bool = None, underlined: bool = None,
                 class_name: str = None):
        if fg is not None:
            self._fg = self._resolve_color(fg, True)
        if bg is not None:
            self._bg = self._resolve_color(bg, True)

        self.blink = blink
        self.bold = bold
        self.crosslined = crosslined
        self.dim = dim
        self.double_underlined = double_underlined
        self.inversed = inversed
        self.italic = italic
        self.overlined = overlined
        self.underlined = underlined
        self.class_name = class_name

        if inherit is not None:
            self._clone_from(inherit)

        if self._fg is None:
            self._fg = NOOP_COLOR
        if self._bg is None:
            self._bg = NOOP_COLOR

    def text(self, text: Any) -> Text:
        return Text(text, self)

    def render(self, text: Any) -> str:
        return self.text(text).render()

    def autopick_fg(self) -> Style:
        """
        Pick ``fg_color`` depending on ``bg_color``. Set ``fg_color`` to
        either 4% gray (almost black) if background is bright, or to 80% gray
        (bright gray) if it is dark. If background is None, do nothing.

        .. todo ::

            check if there is a better algorithm,
            because current thinks text on #000080 should be black

        :return: self
        """
        if self._bg is None or self._bg.hex_value is None:
            return self

        h, s, v = Color.hex_value_to_hsv_channels(self._bg.hex_value)
        if v >= .45:
            self._fg = Colors.RGB_GREY_4
        else:
            self._fg = Colors.RGB_GREY_80
        return self

    def flip(self) -> Style:
        """
        Swap foreground color and background color.
        :return: self
        """
        self._fg, self._bg = self._bg, self._fg
        return self

    # noinspection PyMethodMayBeStatic
    def _resolve_color(self, arg: str|int|Color, nullable: bool) -> Color|None:
        if isinstance(arg, Color):
            return arg

        if isinstance(arg, int):
            return ColorRGB(arg)

        if isinstance(arg, str):
            resolved_color = Colors.resolve(arg)
            if not isinstance(resolved_color, Color):
                raise ValueError(f'Attribute is not valid Color: {resolved_color}')
            return resolved_color

        return None if nullable else NOOP_COLOR

    def _clone_from(self, inherit: Style):
        for attr in list(self.__dict__.keys()) + ['_fg', '_bg']:
            inherit_val = getattr(inherit, attr)
            if getattr(self, attr) is None and inherit_val is not None:
                setattr(self, attr, inherit_val)

    def __repr__(self):
        if self._fg is None or self._bg is None:
            return self.__class__.__name__ + '[uninitialized]'
        props_set = [self.fg.format_value('fg=', 'no fg'),
                     self.bg.format_value('bg=', 'no bg'), ]
        for attr_name in dir(self):
            if not attr_name.startswith('_'):
                attr = getattr(self, attr_name)
                if isinstance(attr, bool) and attr is True:
                    props_set.append(attr_name)

        return self.__class__.__name__ + '[{:s}]'.format(', '.join(props_set))

    @property
    def fg(self) -> Color:
        return self._fg

    @property
    def bg(self) -> Color:
        return self._bg

    @fg.setter
    def fg(self, val: str|int|Color):
        self._fg: Color = self._resolve_color(val, nullable=False)

    @bg.setter
    def bg(self, val: str|int|Color):
        self._bg: Color = self._resolve_color(val, nullable=False)


NOOP_STYLE = Style()
"""
Special style which passes the text 
furthrer without any modifications.
"""


class RendererManager:
    _default: Type[IRenderer] = None

    @classmethod
    def set_up(cls, default_renderer: Type[IRenderer]|None = None):
        """
        Set up renderer preferences. Affects all renderer types.

        :param default_renderer:
            Default renderer to use globally. Passing None will result in library
            default setting restored (`SgrRenderer`).

        >>> RendererManager.set_up(DebugRenderer)
        >>> Text('text', Style(fg='red')).render()
        '|ǝ31|text|ǝ39|'

        >>> NoOpRenderer.render('text',Style(fg='red'))
        'text'
        """
        cls._default = default_renderer or SgrRenderer

    @classmethod
    def get_default(cls) -> Type[IRenderer]:
        """ Get global default renderer type. """
        return cls._default


class IRenderer(metaclass=ABCMeta):
    """ Renderer interface. """

    @classmethod
    @abstractmethod
    def render(cls, text: Any, style: Style = NOOP_STYLE) -> str:
        """
        Apply colors and attributes described in ``style`` argument to
        ``text`` and return the result. Output format depends on renderer's
        class (which defines the implementation).
        """
        raise NotImplementedError


class SgrRenderer(IRenderer):
    """
    Default renderer invoked by `Text._render()`. Transforms `Color` instances
    defined in ``style`` into ANSI control sequence bytes and merges them with
    input string.

    Respects compatibility preferences (see `RendererManager.set_up()`) and
    maps RGB colors to closest *indexed* colors if terminal doesn't support
    RGB output. In case terminal doesn't support even 256 colors, falls back
    to 16-color pallete and picks closest counterparts again the same way.

    Type of output ``SequenceSGR`` depends on type of `Color` variables in
    ``style`` argument. Keeping all that in mind, let's summarize:

    1. `ColorRGB` can be rendered as True Color sequence, 256-color sequence
       or 16-color sequence depending on compatibility settings.
    2. `ColorIndexed256` can be rendered as 256-color sequence or 16-color
       sequence.
    3. `ColorIndexed16` can be rendered as 16-color sequence.
    4. Nothing of the above will happen and all Colors will be discarded
       completely if output is not a terminal emulator or if the developer
       explicitly set up the renderer to do so (**force_styles** = False).

    >>> SgrRenderer.render('text', Style(fg='red', bold=True))
    '\\x1b[1;31mtext\\x1b[22;39m'
    """
    _force_styles: bool|None = False
    _compatibility_256_colors: bool = False
    _compatibility_16_colors: bool = False

    @classmethod
    def set_up(cls, force_styles: bool|None = False,
               compatibility_256_colors: bool = False,
               compatibility_16_colors: bool = False):
        """
        Set up renderer preferences. Affects all renderer types.

        .. todo ::
            Rewrite this part. Default should be *256* OR *RGB* if COLORTERM is either
            ``truecolor`` or ``24bit``. `setup()` overrides this, of course.

        :param force_styles:

            * If set to *None*, all renderers will pass input text through themselves
              without any changes (i.e. no colors and attributes will be applied).
            * If set to *True*, renderers will always apply the formatting regardless
              of other internal rules and algorithms.
            * If set to *False* [default], the final decision will be made
              by every renderer independently, based on their own algorithms.

        :param compatibility_256_colors:

            Disable *RGB* (or True Color) output mode. *256-color* sequences will
            be printed out instead of disabled ones. Useful when combined with
            ``curses`` -- that way you can check the terminal capabilities from the
            inside of that terminal and switch to different output mode at once.

        :param compatibility_16_colors:

            Disable *256-color* output mode and default *16-color* sequences instead.
            If this setting is set to *True*, the value of ``compatibility_256_colors``
            will be ignored completely.

        """
        cls._force_styles = force_styles
        cls._compatibility_256_colors = compatibility_256_colors
        cls._compatibility_16_colors = compatibility_16_colors

    @classmethod
    def render(cls, text: Any, style: Style = NOOP_STYLE):
        opening_seq = cls._render_attributes(style, squash=True) + \
                      cls._render_color(style.fg, False) + \
                      cls._render_color(style.bg, True)

        # in case there are line breaks -- split text to lines and apply
        # SGRs for each line separately. it increases the chances that style
        # will be correctly displayed regardless of implementation details of
        # user's pager, multiplexer, terminal emulator etc.
        rendered_text = ''
        for line in str(text).splitlines(keepends=True):
            rendered_text += Span(opening_seq).wrap(line)
        return rendered_text

    @classmethod
    def _render_attributes(cls, style: Style, squash: bool) -> List[SequenceSGR]|SequenceSGR:
        if not cls.is_sgr_usage_allowed():
            return NOOP_SEQ if squash else [NOOP_SEQ]

        result = []
        if style.blink:             result += [Seqs.BLINK_SLOW]
        if style.bold:              result += [Seqs.BOLD]
        if style.crosslined:        result += [Seqs.CROSSLINED]
        if style.dim:               result += [Seqs.DIM]
        if style.double_underlined: result += [Seqs.DOUBLE_UNDERLINED]
        if style.inversed:          result += [Seqs.INVERSED]
        if style.italic:            result += [Seqs.ITALIC]
        if style.overlined:         result += [Seqs.OVERLINED]
        if style.underlined:        result += [Seqs.UNDERLINED]

        if squash:
            return reduce(lambda p, c: p + c, result, NOOP_SEQ)
        return result

    @classmethod
    def _render_color(cls, color: Color, bg: bool) -> SequenceSGR:
        hex_value = color.hex_value

        if not cls.is_sgr_usage_allowed() or hex_value is None:
            return NOOP_SEQ

        if isinstance(color, ColorRGB):
            if cls._compatibility_16_colors:
                return ColorIndexed16.find_closest(hex_value).to_sgr(bg=bg)
            if cls._compatibility_256_colors:
                return ColorIndexed256.find_closest(hex_value).to_sgr(bg=bg)
            return color.to_sgr(bg=bg)

        elif isinstance(color, ColorIndexed256):
            if cls._compatibility_16_colors:
                return ColorIndexed16.find_closest(hex_value).to_sgr(bg=bg)
            return color.to_sgr(bg=bg)

        elif isinstance(color, ColorIndexed16):
            return color.to_sgr(bg=bg)

        raise NotImplementedError(f'Unknown Color inhertior {color!s}')

    @classmethod
    def is_sgr_usage_allowed(cls) -> bool:
        if cls._force_styles is True:
            return True
        if cls._force_styles is None:
            return False
        return sys.stdout.isatty()


class TmuxRenderer(SgrRenderer):
    """
    tmux
    """

    SGR_TO_TMUX_MAP = {
        NOOP_SEQ:           '',
        Seqs.RESET:         'default',

        Seqs.BOLD:          'bold',
        Seqs.DIM:           'dim',
        Seqs.ITALIC:        'italics',
        Seqs.UNDERLINED:    'underscore',
        Seqs.BLINK_SLOW:    'blink',
        Seqs.BLINK_FAST:    'blink',
        Seqs.BLINK_DEFAULT: 'blink',
        Seqs.INVERSED:      'reverse',
        Seqs.HIDDEN:        'hidden',
        Seqs.CROSSLINED:    'strikethrough',
        Seqs.DOUBLE_UNDERLINED: 'double-underscore',
        Seqs.OVERLINED:     'overline',

        Seqs.BOLD_DIM_OFF:   'nobold nodim',
        Seqs.ITALIC_OFF:     'noitalics',
        Seqs.UNDERLINED_OFF: 'nounderscore',
        Seqs.BLINK_OFF:      'noblink',
        Seqs.INVERSED_OFF:   'noreverse',
        Seqs.HIDDEN_OFF:     'nohidden',
        Seqs.CROSSLINED_OFF: 'nostrikethrough',
        Seqs.OVERLINED_OFF:  'nooverline',

        Seqs.BLACK:     'fg=black',
        Seqs.RED:       'fg=red',
        Seqs.GREEN:     'fg=green',
        Seqs.YELLOW:    'fg=yellow',
        Seqs.BLUE:      'fg=blue',
        Seqs.MAGENTA:   'fg=magenta',
        Seqs.CYAN:      'fg=cyan',
        Seqs.WHITE:     'fg=white',
        Seqs.COLOR_OFF: 'fg=default',

        Seqs.BG_BLACK:     'bg=black',
        Seqs.BG_RED:       'bg=red',
        Seqs.BG_GREEN:     'bg=green',
        Seqs.BG_YELLOW:    'bg=yellow',
        Seqs.BG_BLUE:      'bg=blue',
        Seqs.BG_MAGENTA:   'bg=magenta',
        Seqs.BG_CYAN:      'bg=cyan',
        Seqs.BG_WHITE:     'bg=white',
        Seqs.BG_COLOR_OFF: 'bg=default',

        Seqs.GRAY:       'fg=brightblack',
        Seqs.HI_RED:     'fg=brightred',
        Seqs.HI_GREEN:   'fg=brightgreen',
        Seqs.HI_YELLOW:  'fg=brightyellow',
        Seqs.HI_BLUE:    'fg=brightblue',
        Seqs.HI_MAGENTA: 'fg=brightmagenta',
        Seqs.HI_CYAN:    'fg=brightcyan',
        Seqs.HI_WHITE:   'fg=brightwhite',

        Seqs.BG_GRAY:       'bg=brightblack',
        Seqs.BG_HI_RED:     'bg=brightred',
        Seqs.BG_HI_GREEN:   'bg=brightgreen',
        Seqs.BG_HI_YELLOW:  'bg=brightyellow',
        Seqs.BG_HI_BLUE:    'bg=brightblue',
        Seqs.BG_HI_MAGENTA: 'bg=brightmagenta',
        Seqs.BG_HI_CYAN:    'bg=brightcyan',
        Seqs.BG_HI_WHITE:   'bg=brightwhite',
    }

    @classmethod
    def render(cls, text: Any, style: Style = NOOP_STYLE):
        opening_sgrs = [
            *cls._render_attributes(style, False),
            cls._render_color(style.fg, False),
            cls._render_color(style.bg, True),
        ]
        opening_tmux_style = cls._sgr_to_tmux_style(*opening_sgrs)
        closing_tmux_style = ''.join(
            cls._sgr_to_tmux_style(Span(sgr).closing_seq) for sgr in opening_sgrs
        )

        rendered_text = ''
        for line in str(text).splitlines(keepends=True):
            rendered_text += opening_tmux_style + line + closing_tmux_style
        return rendered_text

    @classmethod
    def _sgr_to_tmux_style(cls, *sgrs: SequenceSGR) -> str:
        result = ''
        for sgr in sgrs:
            if sgr.is_color_extended:
                target = 'fg'
                if sgr.params[0] == IntCodes.BG_COLOR_EXTENDED:
                    target = 'bg'

                if sgr.params[1] == IntCodes.EXTENDED_MODE_256:
                    color = 'color{}'.format(sgr.params[2])
                elif sgr.params[1] == IntCodes.EXTENDED_MODE_RGB:
                    color = '#{:06x}'.format(
                        ColorRGB.rgb_channels_to_hex_value(*sgr.params[2:])
                    )
                else:
                    raise ValueError(f"Unknown SGR param #2 (idx 1): {sgr!r}")

                result += f'#[{target}={color}]'
                continue

            tmux_style_decl = cls.SGR_TO_TMUX_MAP.get(sgr)
            if tmux_style_decl is None:
                raise LogicError(f'No tmux definiton is present for {sgr!r}')
            if len(tmux_style_decl) > 0:
                result += f'#[{tmux_style_decl}]'
        return result


class NoOpRenderer(IRenderer):
    """
    Special renderer type that does nothing with the input string and just
    returns it as is. That's true only when it _is_ a str beforehand;
    otherwise argument will be casted to str and then returned.

    >>> NoOpRenderer.render('text', Style(fg='red', bold=True))
    'text'
    """

    @classmethod
    def render(cls, text: Any, style: Style = NOOP_STYLE) -> str:
        return str(text)


class HtmlRenderer(IRenderer):
    """
    html

    >>> HtmlRenderer.render('text',Style(fg='red', bold=True))
    '<span style="color: #800000; font-weight: 700">text</span>'
    """

    DEFAULT_ATTRS = ['color', 'background-color', 'font-weight', 'font-style',
                     'text-decoration', 'border', 'filter', ]

    @classmethod
    def render(cls, text: Any, style: Style = NOOP_STYLE) -> str:
        span_styles: Dict[str, Set[str]] = dict()
        for attr in cls._get_default_attrs():
            span_styles[attr] = set()

        if style.fg.hex_value is not None:
            span_styles['color'].add(style.fg.format_value("#"))
        if style.bg.hex_value is not None:
            span_styles['background-color'].add(style.bg.format_value("#"))

        if style.blink:  # modern browsers doesn't support it without shit piled up
            span_styles['border'].update(('1px', 'dotted'))
        if style.bold:
            span_styles['font-weight'].add('700')
        if style.crosslined:
            span_styles['text-decoration'].add('line-through')
        if style.dim:
            span_styles['filter'].update(('saturate(0.5)', 'brightness(0.75)'))
        if style.double_underlined:
            span_styles['text-decoration'].update(('underline', 'double'))
        if style.inversed:
            span_styles['color'], span_styles['background-color'] = \
                span_styles['background-color'], span_styles['color']
        if style.italic:
            span_styles['font-style'].add('italic')
        if style.overlined:
            span_styles['text-decoration'].add('overline')
        if style.underlined:
            span_styles['text-decoration'].add('underline')

        span_class_str = '' if style.class_name is None else f' class="{style.class_name}"'
        span_style_str = '; '.join(f"{k}: {' '.join(v)}"
                                   for k, v in span_styles.items()
                                   if len(v) > 0)
        return f'<span{span_class_str} style="{span_style_str}">'+str(text)+'</span>'  # @TODO  # attribues

    @classmethod
    def _get_default_attrs(cls) -> List[str]:
        return cls.DEFAULT_ATTRS


class DebugRenderer(SgrRenderer):
    """
    DebugRenderer

    >>> DebugRenderer.render('text',Style(fg='red', bold=True))
    '|ǝ1;31|text|ǝ22;39|'
    """

    @classmethod
    def render(cls, text: Any, style: Style = NOOP_STYLE) -> str:
        return ReplaceSGR(r'|ǝ\3|').apply(super().render(str(text), style))

    @classmethod
    def is_sgr_usage_allowed(cls) -> bool:
        return True


RendererManager.set_up()


class Styles(Registry[Style]):
    """
    Some ready-to-use styles. Can be used as examples.

    This registry has unique keys in comparsion with other ones (`Seqs` / `Spans` /
    `IntCodes`),
    Therefore there is no risk of key/value duplication and all presets can be listed
    in the initial place -- at API docs page directly.
    """

    WARNING = Style(fg=Colors.YELLOW)
    WARNING_LABEL = Style(WARNING, bold=True)
    WARNING_ACCENT = Style(fg=Colors.HI_YELLOW)

    ERROR = Style(fg=Colors.RED)
    ERROR_LABEL = Style(ERROR, bold=True)
    ERROR_ACCENT = Style(fg=Colors.HI_RED)

    CRITICAL = Style(bg=Colors.HI_RED, fg=Colors.HI_WHITE)
    CRITICAL_LABEL = Style(CRITICAL, bold=True)
    CRITICAL_ACCENT = Style(CRITICAL, bold=True, blink=True)


def distribute_padded(values: List, max_len: int, pad_before: bool = False,
                      pad_after: bool = False, ) -> str:
    if pad_before:
        values.insert(0, '')
    if pad_after:
        values.append('')

    values_amount = len(values)
    gapes_amount = values_amount - 1
    values_len = sum(len(v) for v in values)
    spaces_amount = max_len - values_len
    if spaces_amount < gapes_amount:
        raise ValueError(f'There is not enough space for all values with padding')

    result = ''
    for value_idx, value in enumerate(values):
        gape_len = spaces_amount // (gapes_amount or 1)  # for last value
        result += value + ' ' * gape_len
        gapes_amount -= 1
        spaces_amount -= gape_len

    return result



# -----------------------------------------------------------------------------
#  pytermor [ANSI formatted terminal output toolset]
#  (c) 2022. A. Shavykin <0.delameter@gmail.com>
# -----------------------------------------------------------------------------
"""
Module for time difference formatting (e.g. "4 days 15 hours", "8h 59m").

Supports several output lengths and can be customized even more.

.. testsetup:: *

    from pytermor.util.time_delta import format_time_delta

"""
from __future__ import annotations

from dataclasses import dataclass
from math import floor, trunc, isclose
from typing import List, Dict

from .stdlib_ext import rjust_sgr


def format_time_delta(seconds: float, max_len: int = None) -> str:
    """
    Format time delta using suitable format (which depends on
    ``max_len`` argument). Key feature of this formatter is
    ability to combine two units and display them simultaneously,
    e.g. return "3h 48min" instead of "228 mins" or "3 hours",

    There are predefined formatters with output length of 3, 4,
    6 and 10 characters. Therefore, you can pass in any value
    from 3 inclusive and it's guarenteed that result's length
    will be less or equal to required length. If `max_len` is
    omitted, longest registred formatter will be used.

    >>> format_time_delta(10, 3)
    '10s'
    >>> format_time_delta(10, 6)
    '10 sec'
    >>> format_time_delta(15350, 4)
    '4 h'
    >>> format_time_delta(15350)
    '4h 15min'

    :param seconds: Value to format
    :param max_len: Maximum output string length (total)
    :return:        Formatted string
    """
    if max_len is None:
        formatter = registry.get_longest()
    else:
        formatter = registry.find_matching(max_len)

    if formatter is None:
        raise ValueError(f'No settings defined for max length = {max_len} (or less)')

    return formatter.format(seconds)


class TimeDeltaFormatter:
    """
    Formatter for time intervals. Key feature of this formatter is
    ability to combine two units and display them simultaneously,
    e.g. return "3h 48min" instead of "228 mins" or "3 hours", etc.

    You can create your own formatters if you need fine tuning of the
    output and customization. If that's not the case, there is a
    facade method :meth:`format_time_delta()` which will select appropriate
    formatter automatically.

    Example output::

        "10 secs", "5 mins", "4h 15min", "5d 22h"

    """
    def __init__(self, units: List[TimeUnit], allow_negative: bool, unit_separator: str = None,
                 plural_suffix: str = None,  overflow_msg: str = 'OVERFLOW'):
        self._units = units
        self._allow_negative = allow_negative
        self._unit_separator = unit_separator
        self._plural_suffix = plural_suffix
        self._overflow_msg = overflow_msg

        self._max_len = self._compute_max_len()

    @property
    def max_len(self) -> int:
        """
        This property cannot be set manually, it is
        computed on initialization automatically.

        :return: Maximum possible output string length.
        """
        return self._max_len

    def format(self, seconds: float, always_max_len: bool = False) -> str:
        """
        Pretty-print difference between two moments in time.

        :param seconds: Input value.
        :param always_max_len:
                         If result string is less than `max_len` it will be returned
                         as is, unless this flag is set to *True*. In that case output
                         string will be padded with spaces on the left side so that
                         resulting length would be always equal to maximum length.
        :return:  Formatted string.
        """
        result = self.format_raw(seconds)
        if result is None:
            result = self._overflow_msg[:self.max_len]

        if always_max_len:
            result = rjust_sgr(result, self._max_len)

        return result

    def format_raw(self, seconds: float) -> str|None:
        """
        Pretty-print difference between two moments in time, do not replace
        the output with "OVERFLOW" warning message.

        :param seconds: Input value.
        :return:        Formatted string or *None* on overflow (if input
                        value is too big for the current formatter to handle).
        """
        num = abs(seconds)
        unit_idx = 0
        prev_frac = ''

        negative = self._allow_negative and seconds < 0
        sign = '-' if negative else ''
        result = None

        while result is None and unit_idx < len(self._units):
            unit = self._units[unit_idx]
            if unit.overflow_afer and num > unit.overflow_afer:
                if not self._max_len:  # max len is being computed now
                    raise RecursionError()
                return None

            unit_name = unit.name
            unit_name_suffixed = unit_name
            if self._plural_suffix and trunc(num) != 1:
                unit_name_suffixed += self._plural_suffix

            short_unit_name = unit_name[0]
            if unit.custom_short:
                short_unit_name = unit.custom_short

            next_unit_ratio = unit.in_next
            unit_separator = self._unit_separator or ''

            if abs(num) < 1:
                if negative:
                    result = f'~0{unit_separator}{unit_name_suffixed:s}'
                elif isclose(num, 0, abs_tol=1e-03):
                    result = f'0{unit_separator}{unit_name_suffixed:s}'
                else:
                    result = f'<1{unit_separator}{unit_name:s}'

            elif unit.collapsible_after is not None and num < unit.collapsible_after:
                result = f'{sign}{floor(num):d}{short_unit_name:s}{unit_separator}{prev_frac:<s}'

            elif not next_unit_ratio or num < next_unit_ratio:
                result = f'{sign}{floor(num):d}{unit_separator}{unit_name_suffixed:s}'

            else:
                next_num = floor(num / next_unit_ratio)
                prev_frac = '{:d}{:s}'.format(floor(num - (next_num * next_unit_ratio)), short_unit_name)
                num = next_num
                unit_idx += 1
                continue

        return result or ''

    def _compute_max_len(self) -> int:
        max_len = 0
        coef = 1.00

        for unit in self._units:
            test_val = unit.in_next
            if not test_val:
                test_val = unit.overflow_afer
            if not test_val:
                continue
            test_val_seconds = coef * (test_val - 1) * (-1 if self._allow_negative else 1)

            try:
                max_len_unit = self.format_raw(test_val_seconds)
            except RecursionError:
                continue

            max_len = max(max_len, len(max_len_unit))
            coef *= unit.in_next or unit.overflow_afer

        return max_len


@dataclass(frozen=True)
class TimeUnit:
    name: str
    in_next: int = None             # how many current units equal to the (one) next unit
    custom_short: str = None
    collapsible_after: int = None   # min threshold for double-delta to become regular
    overflow_afer: int = None       # max threshold


class _TimeDeltaFormatterRegistry:
    """
    Simple registry for storing formatters and selecting
    the suitable one by max output length.
    """
    def __init__(self):
        self._formatters: Dict[int, TimeDeltaFormatter] = dict()

    def register(self, *formatters: TimeDeltaFormatter):
        for formatter in formatters:
            self._formatters[formatter.max_len] = formatter

    def find_matching(self, max_len: int) -> TimeDeltaFormatter | None:
        exact_match = self.get_by_max_len(max_len)
        if exact_match is not None:
            return exact_match

        suitable_max_len_list = sorted(
            [key for key in self._formatters.keys() if key <= max_len],
            key=lambda k: k,
            reverse=True,
        )
        if len(suitable_max_len_list) == 0:
            return None
        return self.get_by_max_len(suitable_max_len_list[0])

    def get_by_max_len(self, max_len: int) -> TimeDeltaFormatter | None:
        return self._formatters.get(max_len, None)

    def get_shortest(self) -> _TimeDeltaFormatterRegistry | None:
        return self._formatters.get(min(self._formatters.keys() or [None]))

    def get_longest(self) -> _TimeDeltaFormatterRegistry | None:
        return self._formatters.get(max(self._formatters.keys() or [None]))


# ---------------------------------------------------------------------------
# Preset formatters
# ---------------------------------------------------------------------------

registry = _TimeDeltaFormatterRegistry()
registry.register(
    TimeDeltaFormatter([
        TimeUnit('s', 60),
        TimeUnit('m', 60),
        TimeUnit('h', 24),
        TimeUnit('d', overflow_afer=99),
    ], allow_negative=False,
        unit_separator=None,
        plural_suffix=None,
        overflow_msg='ERR',
    ),

    TimeDeltaFormatter([
        TimeUnit('s', 60),
        TimeUnit('m', 60),
        TimeUnit('h', 24),
        TimeUnit('d', 30),
        TimeUnit('M', 12),
        TimeUnit('y', overflow_afer=99),
    ], allow_negative=False,
        unit_separator=' ',
        plural_suffix=None,
        overflow_msg='ERRO',
    ),

    TimeDeltaFormatter([
        TimeUnit('sec', 60),
        TimeUnit('min', 60),
        TimeUnit('hr', 24, collapsible_after=10),
        TimeUnit('day', 30, collapsible_after=10),
        TimeUnit('mon', 12),
        TimeUnit('yr', overflow_afer=99),
    ], allow_negative=False,
        unit_separator=' ',
        plural_suffix=None,
    ),

    TimeDeltaFormatter([
        TimeUnit('sec', 60),
        TimeUnit('min', 60, custom_short='min'),
        TimeUnit('hour', 24, collapsible_after=24),
        TimeUnit('day', 30, collapsible_after=10),
        TimeUnit('month', 12),
        TimeUnit('year', overflow_afer=999),
    ], allow_negative=True,
        unit_separator=' ',
        plural_suffix='s',
    ),
)



# -----------------------------------------------------------------------------
# es7s/kolombos [Escape sequences and control characters visualiser]
# (C) 2022 A. Shavykin <0.delameter@gmail.com>
# -----------------------------------------------------------------------------
from __future__ import annotations

from typing import TypeVar, Generic, Dict

from pytermor import SequenceSGR

from kolombos.byteio import ByteIOMode

KT = TypeVar('KT')
VT = TypeVar('VT')


class PartialOverride(Generic[KT, VT]):
    """Class used to reduce overhead when dealing with abstractions with
    lots of different parameter values, which are *almost always* equal
    to some default value.

    """
    def __init__(self, default: VT, override: Dict[KT, VT] = None):
        self._default: VT = default
        self._override: Dict[KT, VT]|None = override

    def get(self, *keys: KT) -> VT:
        """Returns first found override value, or default value if nothing
         found. Iterates through arguments from left to right (i.e. first
         `key` has the highest priority).

        """
        if self._override:
            for key in keys:
                if key in self._override.keys():
                    return self._override[key]
        return self._default

    def set(self, key: KT, value: VT):
        if not self._override:
            self._override = dict()
        self._override[key] = value

    def has_key(self, key: KT) -> bool:
        if not self._override:
            return False
        return key in self._override.keys()


class OpeningSeqPOV(PartialOverride[ByteIOMode, SequenceSGR]):
    pass


class LabelPOV(PartialOverride[ByteIOMode, str]):
    pass



# -----------------------------------------------------------------------------
# es7s/kolombos [Escape sequences and control characters visualiser]
# (C) 2022 A. Shavykin <0.delameter@gmail.com>
# -----------------------------------------------------------------------------
from __future__ import annotations

from pytermor import SequenceSGR

from . import Template
from .. import CharClass, OpeningSeqPOV, LabelPOV


class Utf8SequenceTemplate(Template):
    DECODED_LEFT_FILL_CHAR = '_'

    def __init__(self, opening_seq: SequenceSGR | OpeningSeqPOV, label: str | LabelPOV = ''):
        super().__init__(CharClass.UTF_8_SEQ, opening_seq, label)

    def _process(self, raw: bytes) -> str:
        default = lambda: super(type(self), self)._process(raw)

        if self._display_mode.is_ignored:
            return default()

        if self._read_mode.is_text or self._decode:
            decoded = raw.decode('utf8', errors='replace')
            if self._read_mode.is_binary:
                if len(decoded) < len(raw):
                    decoded = decoded.rjust(len(raw), self.DECODED_LEFT_FILL_CHAR)
                elif len(decoded) > len(raw):
                    decoded = decoded[:len(raw)]
            return decoded

        return default()



# -----------------------------------------------------------------------------
# es7s/kolombos [Escape sequences and control characters visualiser]
# (C) 2022 A. Shavykin <0.delameter@gmail.com>
# -----------------------------------------------------------------------------
from __future__ import annotations

from . import AbstractRunner, LegendRunner, VersionRunner, ByteIoRunner
from .. import ArgumentError
from ..settings import SettingsManager


class RunnerFactory:
    @staticmethod
    def create() -> AbstractRunner:
        if SettingsManager.app_settings.legend:
            return LegendRunner()
        elif SettingsManager.app_settings.version:
            return VersionRunner()
        elif SettingsManager.app_settings.binary or SettingsManager.app_settings.text:
            return ByteIoRunner()
        raise ArgumentError('No mode specified')



# -----------------------------------------------------------------------------
#  pytermor [ANSI formatted terminal output toolset]
#  (c) 2022. A. Shavykin <0.delameter@gmail.com>
# -----------------------------------------------------------------------------
"""
Module contains definitions for low-level ANSI escape sequences handling.

The key difference beetween ``Spans`` and ``Sequences`` is that sequence can
*open* text style and also *close*, or terminate it. As for ``Spans`` -- they
always do both; typical use-case of `Span` is to wrap some text in opening SGR
and closing one.

Each variable in `Seqs` and `Spans` below is a valid argument
for :class:`.Span` and :class:`.SequenceSGR` default constructors; furthermore,
it can be passed in a string form (case-insensitive):

.. testsetup:: *

    from pytermor.ansi import SequenceSGR, Spans, Span, Seqs, IntCodes, NOOP_SPAN, NOOP_SEQ

>>> Span('BG_GREEN')
Span[SGR[42], SGR[49]]

>>> Span(Seqs.BG_GREEN, Seqs.UNDERLINED)
Span[SGR[42;4], SGR[49;24]]

"""
from __future__ import annotations

from abc import ABCMeta, abstractmethod
from copy import copy
from typing import List, Any, Dict, Tuple, Sized

from .common import Registry


class Sequence(Sized, metaclass=ABCMeta):
    """
    Abstract ancestor of all escape sequences.
    """
    def __init__(self, *params: int):
        self._params: List[int] = [max(0, int(p)) for p in params]

    @abstractmethod
    def assemble(self) -> str:
        """
        Build up actual byte sequence and return
        as an ASCII-encoded string.
        """
        raise NotImplementedError

    @property
    def params(self) -> List[int]:
        """ Return internal params as array. """
        return self._params

    @classmethod
    @abstractmethod
    def _short_class_name(cls): raise NotImplementedError

    def __str__(self) -> str:
        return self.assemble()

    def __len__(self) -> int:
        return 0

    def __eq__(self, other: Sequence):
        if type(self) != type(other):
            return False
        return self._params == other._params

    def __repr__(self):
        params = ";".join([str(p) for p in self._params])
        if len(self._params) == 0:
            params = '~'
        return f'{self._short_class_name()}[{params}]'


class SequenceCSI(Sequence, metaclass=ABCMeta):
    """
    Abstract class representing CSI-type ANSI escape sequence. All subtypes
    of this sequence start with :kbd:`\\e[`.
    """
    _CONTROL_CHARACTER = '\x1b'
    _INTRODUCER = '['
    _SEPARATOR = ';'

    def __init__(self, *params: int):
        super(SequenceCSI, self).__init__(*params)

    @classmethod
    def regexp(cls) -> str:
        return f'\\x1b\\[[0-9;]*{cls._terminator()}'

    @classmethod
    @abstractmethod
    def _terminator(cls) -> str: raise NotImplementedError


class SequenceSGR(SequenceCSI, metaclass=ABCMeta):
    """
    Class representing SGR-type escape sequence with varying amount of parameters.

    `SequenceSGR` with zero params was specifically implemented to
    translate into empty string and not into :kbd:`\\e[m`, which would have
    made sense, but also would be very entangling, as this sequence is
    equivalent of :kbd:`\\e[0m` -- hard reset sequence. The empty-string-sequence
    is predefined as `NOOP_SEQ`.

    It's possible to add of one SGR sequence to another:

    >>> SequenceSGR(31) + SequenceSGR(1) == SequenceSGR(31, 1)
    True

    """
    _TERMINATOR = 'm'

    def __init__(self, *args: str|int|SequenceSGR):
        """
        Create new `SequenceSGR` with specified ``args`` as params.

        Resulting sequence param order is same as an argument order.

        Each sequence param can be specified as:
          - string key (name of any constant defined in `IntCodes`, case-insensitive)
          - integer param value (``IntCodes`` values)
          - existing ``SequenceSGR`` instance (params will be extracted).

        >>> SequenceSGR('yellow', 'bold')
        SGR[33;1]
        >>> SequenceSGR(91, 7)
        SGR[91;7]
        >>> SequenceSGR(IntCodes.HI_CYAN, IntCodes.UNDERLINED)
        SGR[96;4]
        """
        result: List[int] = []

        for arg in args:
            if isinstance(arg, str):
                resolved_param = IntCodes.resolve(arg)
                if not isinstance(resolved_param, int):
                    raise ValueError(f'Attribute is not valid SGR param: {resolved_param}')
                result.append(resolved_param)
            elif isinstance(arg, int):
                result.append(arg)
            elif isinstance(arg, SequenceSGR):
                result.extend(arg.params)
            else:
                raise TypeError(f'Invalid argument type: {arg!r})')

        super().__init__(*result)

    @classmethod
    def init_color_indexed(cls, idx: int, bg: bool = False) -> SequenceSGR:
        """
        Wrapper for creation of `SequenceSGR` that sets foreground
        (or background) to one of 256-color pallete value.

        :param idx:  Index of the color in the pallete, 0 -- 255.
        :param bg:    Set to *True* to change the background color
                      (default is foreground).
        :return:      `SequenceSGR` with required params.
        """

        cls._validate_extended_color(idx)
        key_code = IntCodes.BG_COLOR_EXTENDED if bg else IntCodes.COLOR_EXTENDED
        return SequenceSGR(key_code, IntCodes.EXTENDED_MODE_256, idx)

    @classmethod
    def init_color_rgb(cls, r: int, g: int, b: int, bg: bool = False) -> SequenceSGR:
        """
        Wrapper for creation of `SequenceSGR` operating in True Color mode (16M).
        Valid values for *r*, *g* and *b* are in range [0; 255]. This range
        linearly translates into [``0x00``; ``0xFF``] for each channel. The result
        value is composed as ``#RRGGBB``. For example, sequence with color of
        ``#FF3300`` can be created with::

            SequenceSGR.init_color_rgb(255, 51, 0)

        :param r:  Red channel value, 0 -- 255.
        :param g:  Blue channel value, 0 -- 255.
        :param b:  Green channel value, 0 -- 255.
        :param bg: Set to *True* to change the background color
                   (default is foreground).
        :return:   `SequenceSGR` with required params.
        """

        [cls._validate_extended_color(color) for color in [r, g, b]]
        key_code = IntCodes.BG_COLOR_EXTENDED if bg else IntCodes.COLOR_EXTENDED
        return SequenceSGR(key_code, IntCodes.EXTENDED_MODE_RGB, r, g, b)

    def assemble(self) -> str:
        if len(self._params) == 0:  # NOOP
            return ''

        params = self._params
        if params == [0]:  # \x1b[0m <=> \x1b[m, saving 1 byte
            params = []

        return (self._CONTROL_CHARACTER +
                self._INTRODUCER +
                self._SEPARATOR.join([str(param) for param in params]) +
                self._TERMINATOR)

    def __hash__(self) -> int:
        return int.from_bytes(self.assemble().encode(), byteorder='big')

    def __add__(self, other: SequenceSGR) -> SequenceSGR:
        self._ensure_sequence(other)
        return SequenceSGR(*self._params, *other._params)

    def __radd__(self, other: SequenceSGR) -> SequenceSGR:
        return other.__add__(self)

    def __iadd__(self, other: SequenceSGR) -> SequenceSGR:
        return self.__add__(other)

    def __eq__(self, other: SequenceSGR):
        if type(self) != type(other):
            return False
        return self._params == other._params

    @property
    def is_color_extended(self) -> bool:
        return len(self.params) >= 3 and \
               (self.params[0] == IntCodes.COLOR_EXTENDED or
                self.params[0] == IntCodes.BG_COLOR_EXTENDED)

    @staticmethod
    def _ensure_sequence(subject: Any):
        if not isinstance(subject, SequenceSGR):
            raise TypeError(f'Expected SequenceSGR, got {type(subject)}')

    @staticmethod
    def _validate_extended_color(value: int):
        if value < 0 or value > 255:
            raise ValueError(
                f'Invalid color value: expected range [0-255], got: {value}')

    @classmethod
    def _terminator(cls) -> str:
        return cls._TERMINATOR

    @classmethod
    def _short_class_name(cls) -> str:
        return 'SGR'


# noinspection PyMethodMayBeStatic
class Span:
    """
    Class consisting of two `SequenceSGR` instances -- the first one, "opener",
    tells the terminal that's it should format subsequent characters as specified,
    and the second one, which reverses the effects of the  first one.
    """
    def __init__(self, *opening_params: str|int|SequenceSGR):
        """
        Create a `Span` with specified control sequence(s) as an opening sequence
        and **automatically compose** second (closing) sequence that will terminate
        attributes defined in the first one while keeping the others (*soft* reset).

        Resulting sequence param order is same as an argument order.

        Each argument can be specified as:
          - string key (name of any constant defined in `IntCodes`, case-insensitive)
          - integer param value (``IntCodes`` values)
          - existing `SequenceSGR` instance (params will be extracted).

        >>> Span('red', 'bold')
        Span[SGR[31;1], SGR[39;22]]
        >>> Span(IntCodes.GREEN)
        Span[SGR[32], SGR[39]]
        >>> Span(93, 4)
        Span[SGR[93;4], SGR[39;24]]
        >>> Span(Seqs.BG_BLACK + Seqs.RED)
        Span[SGR[40;31], SGR[49;39]]

        :param opening_params: string keys, integer codes or existing ``SequenceSGR``
                               instances to build ``Span`` from.
        """
        self._opening_seq = SequenceSGR(*opening_params)
        if len(opening_params) == 0:
            self._closing_seq = SequenceSGR()
            return

        self._closing_seq = _SgrPairityRegistry.get_closing_seq(self._opening_seq)

    @classmethod
    def init_explicit(cls, opening_seq: SequenceSGR = None,
                      closing_seq: SequenceSGR = None,
                      hard_reset_after: bool = False) -> Span:
        """
        Create new `Span` with explicitly specified opening and closing sequences.

        .. note ::
            `closing_seq` gets overwritten with `Seqs.RESET` if ``hard_reset_after`` is *True*.

        :param opening_seq:      Starter sequence, in general determening how `Span` will actually look like.
        :param closing_seq:      Finisher SGR sequence.
        :param hard_reset_after: Terminate *all* formatting after this span.
        """
        instance = Span()
        instance._opening_seq = cls._opt_arg(opening_seq)
        instance._closing_seq = cls._opt_arg(closing_seq)

        if hard_reset_after:
            instance._closing_seq = Seqs.RESET

        return instance

    def wrap(self, text: Any = None) -> str:
        """
        Wrap given ``text`` string with ``SGRs`` defined on initialization -- `opening_seq`
        on the left, `closing_seq` on the right. ``str(text)`` will
        be invoked for all argument types with the exception of *None*,
        which will be replaced with an empty string.

        :param text:  String to wrap.
        :return:      ``text`` enclosed in instance's ``SGRs``, if any.
        """
        result = self._opening_seq.assemble()

        if text is not None:
            result += str(text)

        result += self._closing_seq.assemble()
        return result

    @property
    def opening_str(self) -> str:
        """ Return opening SGR sequence assembled. """
        return self._opening_seq.assemble()

    @property
    def opening_seq(self) -> 'SequenceSGR':
        """ Return opening SGR sequence instance. """
        return self._opening_seq

    @property
    def closing_str(self) -> str:
        """ Return closing SGR sequence assembled. """
        return self._closing_seq.assemble()

    @property
    def closing_seq(self) -> 'SequenceSGR':
        """ Return closing SGR sequence instance. """
        return self._closing_seq

    @classmethod
    def _opt_arg(self, arg: SequenceSGR | None) -> SequenceSGR:
        if not arg:
            return NOOP_SEQ
        return arg

    def __call__(self, text: Any = None) -> str:
        """
        Can be used instead of `wrap()` method.

        >>> Spans.RED('text') == Spans.RED.wrap('text')
        True
        """
        return self.wrap(text)

    def __eq__(self, other: Span) -> bool:
        if not isinstance(other, Span):
            return False
        return self._opening_seq == other._opening_seq and self._closing_seq == other._closing_seq

    def __repr__(self):
        return self.__class__.__name__ + '[{!r}, {!r}]'.format(self._opening_seq, self._closing_seq)


NOOP_SEQ = SequenceSGR()
"""
Special sequence in case you *have to* provide one or another SGR, but do 
not want any control sequences to be actually included in the output. 
``NOOP_SEQ.assemble()`` returns empty string, ``NOOP_SEQ.params`` 
returns empty list.

>>> NOOP_SEQ.assemble()
''
>>> NOOP_SEQ.params
[]
"""

NOOP_SPAN = Span()
"""
Special `Span` in cases where you *have to* select one or 
another `Span`, but do not want any control sequence to be actually included. 

- ``NOOP_SPAN(string)`` or ``NOOP_SPAN.wrap(string)`` returns ``string`` without any modifications;
- ``NOOP_SPAN.opening_str`` and ``NOOP_SPAN.closing_str`` are empty strings;
- ``NOOP_SPAN.opening_seq`` and ``NOOP_SPAN.closing_seq`` both returns `NOOP_SEQ`.

>>> NOOP_SPAN('text')
'text'
>>> NOOP_SPAN.opening_str
''
>>> NOOP_SPAN.opening_seq
SGR[~]
"""


class IntCodes(Registry[int]):
    """
    Complete or almost complete list of reliably working SGR param integer codes.

    Suitable for :class:`.Span` and :class:`.SequenceSGR` default constructors.

    .. attention::
       Registry constants are omitted from API doc pages to improve readability
       and avoid duplication. Summary list of all presets can be found in
       `guide.presets` section of the guide.
    """
    # -- Default attributes and colors --------------------------------------------

    RESET = 0  # hard reset code
    BOLD = 1
    DIM = 2
    ITALIC = 3
    UNDERLINED = 4
    BLINK_SLOW = 5
    BLINK_FAST = 6
    INVERSED = 7
    HIDDEN = 8
    CROSSLINED = 9
    DOUBLE_UNDERLINED = 21
    OVERLINED = 53
    BOLD_DIM_OFF = 22  # there is no separate sequence for disabling either
    ITALIC_OFF = 23               # of BOLD or DIM while keeping the other
    UNDERLINED_OFF = 24
    BLINK_OFF = 25
    INVERSED_OFF = 27
    HIDDEN_OFF = 28
    CROSSLINED_OFF = 29
    COLOR_OFF = 39
    BG_COLOR_OFF = 49
    OVERLINED_OFF = 55

    BLACK = 30
    RED = 31
    GREEN = 32
    YELLOW = 33
    BLUE = 34
    MAGENTA = 35
    CYAN = 36
    WHITE = 37
    COLOR_EXTENDED = 38  # use init_color_indexed() and init_color_rgb() instead

    BG_BLACK = 40
    BG_RED = 41
    BG_GREEN = 42
    BG_YELLOW = 43
    BG_BLUE = 44
    BG_MAGENTA = 45
    BG_CYAN = 46
    BG_WHITE = 47
    BG_COLOR_EXTENDED = 48  # use color_indexed() and color_rgb() instead

    GRAY = 90
    HI_RED = 91
    HI_GREEN = 92
    HI_YELLOW = 93
    HI_BLUE = 94
    HI_MAGENTA = 95
    HI_CYAN = 96
    HI_WHITE = 97

    BG_GRAY = 100
    BG_HI_RED = 101
    BG_HI_GREEN = 102
    BG_HI_YELLOW = 103
    BG_HI_BLUE = 104
    BG_HI_MAGENTA = 105
    BG_HI_CYAN = 106
    BG_HI_WHITE = 107

    # RARELY SUPPORTED (thus excluded)
    # 10-20: font selection
    #    26: proportional spacing
    #    50: disable proportional spacing
    #    51: framed
    #    52: encircled
    #    54: neither framed nor encircled
    # 58-59: underline color
    # 60-65: ideogram attributes
    # 73-75: superscript and subscript

    # -- Default colors lists -----------------------------------------------------

    LIST_COLORS = list(range(30, 39))
    LIST_BG_COLORS = list(range(40, 49))
    LIST_HI_COLORS = list(range(90, 98))
    LIST_BG_HI_COLORS = list(range(100, 108))

    LIST_ALL_COLORS = LIST_COLORS + LIST_BG_COLORS + \
                      LIST_HI_COLORS + LIST_BG_HI_COLORS

    # -- EXTENDED modifiers -------------------------------------------------------

    EXTENDED_MODE_256 = 5
    EXTENDED_MODE_RGB = 2


class Seqs(Registry[SequenceSGR]):
    """
    Registry of sequence presets.

    .. attention::
       Registry constants are omitted from API doc pages to improve readability
       and avoid duplication. Summary list of all presets can be found in
       `guide.presets` section of the guide.
    """

    RESET = SequenceSGR(IntCodes.RESET)
    """
    Hard reset sequence.
    """

    # attributes
    BOLD = SequenceSGR(IntCodes.BOLD)
    DIM = SequenceSGR(IntCodes.DIM)
    ITALIC = SequenceSGR(IntCodes.ITALIC)
    UNDERLINED = SequenceSGR(IntCodes.UNDERLINED)
    BLINK_SLOW = SequenceSGR(IntCodes.BLINK_SLOW)
    BLINK_FAST = SequenceSGR(IntCodes.BLINK_FAST)
    BLINK_DEFAULT = BLINK_SLOW
    INVERSED = SequenceSGR(IntCodes.INVERSED)
    HIDDEN = SequenceSGR(IntCodes.HIDDEN)
    CROSSLINED = SequenceSGR(IntCodes.CROSSLINED)
    DOUBLE_UNDERLINED = SequenceSGR(IntCodes.DOUBLE_UNDERLINED)
    OVERLINED = SequenceSGR(IntCodes.OVERLINED)

    BOLD_DIM_OFF = SequenceSGR(IntCodes.BOLD_DIM_OFF)       # there is no separate sequence for
    ITALIC_OFF = SequenceSGR(IntCodes.ITALIC_OFF)           # disabling either of BOLD or DIM
    UNDERLINED_OFF = SequenceSGR(IntCodes.UNDERLINED_OFF)   # while keeping the other
    BLINK_OFF = SequenceSGR(IntCodes.BLINK_OFF)
    INVERSED_OFF = SequenceSGR(IntCodes.INVERSED_OFF)
    HIDDEN_OFF = SequenceSGR(IntCodes.HIDDEN_OFF)
    CROSSLINED_OFF = SequenceSGR(IntCodes.CROSSLINED_OFF)
    OVERLINED_OFF = SequenceSGR(IntCodes.OVERLINED_OFF)

    # text colors
    BLACK = SequenceSGR(IntCodes.BLACK)
    RED = SequenceSGR(IntCodes.RED)
    GREEN = SequenceSGR(IntCodes.GREEN)
    YELLOW = SequenceSGR(IntCodes.YELLOW)
    BLUE = SequenceSGR(IntCodes.BLUE)
    MAGENTA = SequenceSGR(IntCodes.MAGENTA)
    CYAN = SequenceSGR(IntCodes.CYAN)
    WHITE = SequenceSGR(IntCodes.WHITE)
    # code.COLOR_EXTENDED is handled by color_indexed()
    COLOR_OFF = SequenceSGR(IntCodes.COLOR_OFF)

    # background colors
    BG_BLACK = SequenceSGR(IntCodes.BG_BLACK)
    BG_RED = SequenceSGR(IntCodes.BG_RED)
    BG_GREEN = SequenceSGR(IntCodes.BG_GREEN)
    BG_YELLOW = SequenceSGR(IntCodes.BG_YELLOW)
    BG_BLUE = SequenceSGR(IntCodes.BG_BLUE)
    BG_MAGENTA = SequenceSGR(IntCodes.BG_MAGENTA)
    BG_CYAN = SequenceSGR(IntCodes.BG_CYAN)
    BG_WHITE = SequenceSGR(IntCodes.BG_WHITE)
    # code.BG_COLOR_EXTENDED is handled by color_indexed()
    BG_COLOR_OFF = SequenceSGR(IntCodes.BG_COLOR_OFF)

    # high intensity text colors
    GRAY = SequenceSGR(IntCodes.GRAY)
    HI_RED = SequenceSGR(IntCodes.HI_RED)
    HI_GREEN = SequenceSGR(IntCodes.HI_GREEN)
    HI_YELLOW = SequenceSGR(IntCodes.HI_YELLOW)
    HI_BLUE = SequenceSGR(IntCodes.HI_BLUE)
    HI_MAGENTA = SequenceSGR(IntCodes.HI_MAGENTA)
    HI_CYAN = SequenceSGR(IntCodes.HI_CYAN)
    HI_WHITE = SequenceSGR(IntCodes.HI_WHITE)

    # high intensity background colors
    BG_GRAY = SequenceSGR(IntCodes.BG_GRAY)
    BG_HI_RED = SequenceSGR(IntCodes.BG_HI_RED)
    BG_HI_GREEN = SequenceSGR(IntCodes.BG_HI_GREEN)
    BG_HI_YELLOW = SequenceSGR(IntCodes.BG_HI_YELLOW)
    BG_HI_BLUE = SequenceSGR(IntCodes.BG_HI_BLUE)
    BG_HI_MAGENTA = SequenceSGR(IntCodes.BG_HI_MAGENTA)
    BG_HI_CYAN = SequenceSGR(IntCodes.BG_HI_CYAN)
    BG_HI_WHITE = SequenceSGR(IntCodes.BG_HI_WHITE)


class _SgrPairityRegistry:
    """
    Internal class responsible for correct SGRs termination.
    """
    _code_to_breaker_map: Dict[int|Tuple[int, ...], SequenceSGR] = dict()
    _complex_code_def: Dict[int|Tuple[int, ...], int] = dict()
    _complex_code_max_len: int = 0

    @classmethod
    def __new__(cls, *args, **kwargs):

        _regulars = [(IntCodes.BOLD, IntCodes.BOLD_DIM_OFF),
                     (IntCodes.DIM, IntCodes.BOLD_DIM_OFF),
                     (IntCodes.ITALIC, IntCodes.ITALIC_OFF),
                     (IntCodes.UNDERLINED, IntCodes.UNDERLINED_OFF),
                     (IntCodes.DOUBLE_UNDERLINED, IntCodes.UNDERLINED_OFF),
                     (IntCodes.BLINK_SLOW, IntCodes.BLINK_OFF),
                     (IntCodes.BLINK_FAST, IntCodes.BLINK_OFF),
                     (IntCodes.INVERSED, IntCodes.INVERSED_OFF),
                     (IntCodes.HIDDEN, IntCodes.HIDDEN_OFF),
                     (IntCodes.CROSSLINED, IntCodes.CROSSLINED_OFF),
                     (IntCodes.OVERLINED, IntCodes.OVERLINED_OFF), ]

        for c in _regulars:
            cls.bind_regular(*c)

        for c in [*IntCodes.LIST_COLORS, *IntCodes.LIST_HI_COLORS]:
            cls.bind_regular(c, IntCodes.COLOR_OFF)

        for c in [*IntCodes.LIST_BG_COLORS, *IntCodes.LIST_BG_HI_COLORS]:
            cls.bind_regular(c, IntCodes.BG_COLOR_OFF)

        cls.bind_complex((IntCodes.COLOR_EXTENDED, 5), 1, IntCodes.COLOR_OFF)
        cls.bind_complex((IntCodes.COLOR_EXTENDED, 2), 3, IntCodes.COLOR_OFF)
        cls.bind_complex((IntCodes.BG_COLOR_EXTENDED, 5), 1, IntCodes.BG_COLOR_OFF)
        cls.bind_complex((IntCodes.BG_COLOR_EXTENDED, 2), 3, IntCodes.BG_COLOR_OFF)

    @classmethod
    def bind_regular(cls, starter_code: int|Tuple[int, ...], breaker_code: int):
        if starter_code in cls._code_to_breaker_map:
            raise RuntimeError(f'Conflict: SGR code {starter_code} already '
                               f'has a registered breaker')

        cls._code_to_breaker_map[starter_code] = SequenceSGR(breaker_code)

    @classmethod
    def bind_complex(cls, starter_codes: Tuple[int, ...], param_len: int,
                     breaker_code: int):
        cls.bind_regular(starter_codes, breaker_code)

        if starter_codes in cls._complex_code_def:
            raise RuntimeError(f'Conflict: SGR complex {starter_codes} already '
                               f'has a registered breaker')

        cls._complex_code_def[starter_codes] = param_len
        cls._complex_code_max_len = max(cls._complex_code_max_len,
                                         len(starter_codes) + param_len)

    @classmethod
    def get_closing_seq(cls, opening_seq: SequenceSGR) -> SequenceSGR:
        closing_seq_params: List[int] = []
        opening_params = copy(opening_seq.params)

        while len(opening_params):
            key_params: int|Tuple[int, ...]|None = None

            for complex_len in range(1, min(len(opening_params),
                                            cls._complex_code_max_len + 1)):
                opening_complex_suggestion = tuple(opening_params[:complex_len])

                if opening_complex_suggestion in cls._complex_code_def:
                    key_params = opening_complex_suggestion
                    complex_total_len = (
                        complex_len + cls._complex_code_def[opening_complex_suggestion])
                    opening_params = opening_params[complex_total_len:]
                    break

            if key_params is None:
                key_params = opening_params.pop(0)
            if key_params not in cls._code_to_breaker_map:
                continue

            closing_seq_params.extend(cls._code_to_breaker_map[key_params].params)

        return SequenceSGR(*closing_seq_params)

_SgrPairityRegistry()


class Spans(Registry[Span]):
    """
    Registry of span presets.

    .. attention::
       Registry constants are omitted from API doc pages to improve readability
       and avoid duplication. Summary list of all presets can be found in
       `guide.presets` section of the guide.
    """

    BOLD = Span(IntCodes.BOLD)
    DIM = Span(IntCodes.DIM)
    ITALIC = Span(IntCodes.ITALIC)
    UNDERLINED = Span(IntCodes.UNDERLINED)
    BLINK_SLOW = Span(IntCodes.BLINK_SLOW)
    BLINK_FAST = Span(IntCodes.BLINK_FAST)
    INVERSED = Span(IntCodes.INVERSED)
    HIDDEN = Span(IntCodes.HIDDEN)
    CROSSLINED = Span(IntCodes.CROSSLINED)
    DOUBLE_UNDERLINED = Span(IntCodes.DOUBLE_UNDERLINED)
    OVERLINED = Span(IntCodes.OVERLINED)

    BLACK = Span(IntCodes.BLACK)
    RED = Span(IntCodes.RED)
    GREEN = Span(IntCodes.GREEN)
    YELLOW = Span(IntCodes.YELLOW)
    BLUE = Span(IntCodes.BLUE)
    MAGENTA = Span(IntCodes.MAGENTA)
    CYAN = Span(IntCodes.CYAN)
    WHITE = Span(IntCodes.WHITE)

    GRAY = Span(IntCodes.GRAY)
    HI_RED = Span(IntCodes.HI_RED)
    HI_GREEN = Span(IntCodes.HI_GREEN)
    HI_YELLOW = Span(IntCodes.HI_YELLOW)
    HI_BLUE = Span(IntCodes.HI_BLUE)
    HI_MAGENTA = Span(IntCodes.HI_MAGENTA)
    HI_CYAN = Span(IntCodes.HI_CYAN)
    HI_WHITE = Span(IntCodes.HI_WHITE)

    BG_BLACK = Span(IntCodes.BG_BLACK)
    BG_RED = Span(IntCodes.BG_RED)
    BG_GREEN = Span(IntCodes.BG_GREEN)
    BG_YELLOW = Span(IntCodes.BG_YELLOW)
    BG_BLUE = Span(IntCodes.BG_BLUE)
    BG_MAGENTA = Span(IntCodes.BG_MAGENTA)
    BG_CYAN = Span(IntCodes.BG_CYAN)
    BG_WHITE = Span(IntCodes.BG_WHITE)

    BG_GRAY = Span(IntCodes.BG_GRAY)
    BG_HI_RED = Span(IntCodes.BG_HI_RED)
    BG_HI_GREEN = Span(IntCodes.BG_HI_GREEN)
    BG_HI_YELLOW = Span(IntCodes.BG_HI_YELLOW)
    BG_HI_BLUE = Span(IntCodes.BG_HI_BLUE)
    BG_HI_MAGENTA = Span(IntCodes.BG_HI_MAGENTA)
    BG_HI_CYAN = Span(IntCodes.BG_HI_CYAN)
    BG_HI_WHITE = Span(IntCodes.BG_HI_WHITE)



# -----------------------------------------------------------------------------
# es7s/kolombos [Escape sequences and control characters visualiser]
# (C) 2022 A. Shavykin <0.delameter@gmail.com>
# -----------------------------------------------------------------------------
from ._abstract import AbstractFormatter

from .binary import BinaryFormatter
from .text import TextFormatter

from .factory import FormatterFactory



# -----------------------------------------------------------------------------
# es7s/kolombos [Escape sequences and control characters visualiser]
# (C) 2022 A. Shavykin <0.delameter@gmail.com>
# -----------------------------------------------------------------------------
from __future__ import annotations

from pytermor import SequenceSGR

from . import Chainable


class SequenceRef(Chainable):
    def __init__(self, ref: SequenceSGR):
        self._ref: SequenceSGR = ref

    @property
    def data_len(self) -> int:
        return 0

    @property
    def is_newline(self) -> bool:
        return False

    @property
    def ref(self) -> SequenceSGR:
        return self._ref

    def __eq__(self, other):
        if not isinstance(other, SequenceRef):
            return False
        return self._ref == self._ref

    def __repr__(self):
        return f'{self.__class__.__name__}::{self._ref!r}'


class OneUseSequenceRef(SequenceRef):
    pass


class StartSequenceRef(SequenceRef):
    pass


class StopSequenceRef(SequenceRef):
    pass



# -----------------------------------------------------------------------------
#  pytermor [ANSI formatted terminal output toolset]
#  (c) 2022. A. Shavykin <0.delameter@gmail.com>
# -----------------------------------------------------------------------------
"""
.. testsetup:: *

    from pytermor.util.prefixed_unit import format_si_metric, format_si_binary

"""
from __future__ import annotations

from math import trunc
from typing import List

from .auto_float import format_auto_float


def format_si_metric(value: float, unit: str = 'm') -> str:
    """
    Format ``value`` as meters with SI-prefixes, max
    result length is *6* chars. Base is *1000*. Unit can be
    customized.

    >>> format_si_metric(123.456)
    '123 m'
    >>> format_si_metric(0.331, 'g')
    '331 mg'
    >>> format_si_metric(45200, 'V')
    '45.2 kV'
    >>> format_si_metric(1.26e-9, 'm²')
    '1.26 nm²'

    :param value: Input value (unitless).
    :param unit:  Value unit, printed right after the prefix.
    :return:      Formatted string with SI-prefix if necessary.

    .. versionadded:: 2.0
    """
    return _formatter_si_metric.format(value, unit)


def format_si_binary(value: float, unit: str = 'b') -> str:
    """
    Format ``value`` as binary size (bytes, kbytes, Mbytes), max
    result length is *8* chars. Base is *1024*. Unit can be
    customized.

    >>> format_si_binary(631)
    '631 b'
    >>> format_si_binary(1080)
    '1.055 kb'
    >>> format_si_binary(45200)
    '44.14 kb'
    >>> format_si_binary(1.258 * pow(10, 6), 'bps')
    '1.200 Mbps'

    :param value: Input value in bytes.
    :param unit:  Value unit, printed right after the prefix.
    :return:      Formatted string with SI-prefix if necessary.

    .. versionadded:: 2.0
    """
    return _formatter_si_binary.format(value, unit)


class PrefixedUnitFormatter:
    """
    Formats ``value`` using settings passed to constructor. The main idea of this class
    is to fit into specified string length as much significant digits as it's
    theoretically possible by using multipliers and unit prefixes to
    indicate them.

    You can create your own formatters if you need fine tuning of the
    output and customization. If that's not the case, there are facade
    methods :meth:`format_si_metric()` and :meth:`format_si_binary()`,
    which will invoke predefined formatters and doesn't require setting up.

    .. todo :: params

    :param prefix_zero_idx:
            Index of prefix which will be used as default, i.e. without multiplying coefficients.

    .. versionadded:: 1.7
    """
    def __init__(self, max_value_len: int, truncate_frac: bool = False, unit: str = None, unit_separator: str = None,
                 mcoef: float = 1000.0, prefixes: List[str | None] = None, prefix_zero_idx: int = None):
        self._max_value_len: int = max_value_len
        self._truncate_frac: bool = truncate_frac
        self._unit: str = unit or ''
        self._unit_separator: str = unit_separator or ''
        self._mcoef: float = mcoef
        self._prefixes: List[str | None] = prefixes or []
        self._prefix_zero_idx: int = prefix_zero_idx or 0

    @property
    def max_len(self) -> int:
        """
        :return: Maximum length of the result. Note that constructor argument
                 is `max_value_len`, which is different parameter.
        """
        result = self._max_value_len
        result += len(self._unit_separator)
        result += len(self._unit)
        result += max([len(p) for p in self._prefixes if p])
        return result

    def format(self, value: float, unit: str = None) -> str:
        """
        :param value:  Input value
        :param unit:   Unit override
        :return:       Formatted value
        """
        unit_idx = self._prefix_zero_idx

        while 0 <= unit_idx < len(self._prefixes):
            if 0.0 < abs(value) <= 1 / self._mcoef:
                value *= self._mcoef
                unit_idx -= 1
                if abs(value) <= 1:
                    continue
            elif abs(value) >= self._mcoef:
                value /= self._mcoef
                unit_idx += 1
                continue

            unit_full = (self._prefixes[unit_idx] or '') + \
                        (unit or self._unit)

            if self._truncate_frac and unit_idx == self._prefix_zero_idx:
                num_str = f'{trunc(value)!s:.{self._max_value_len}s}'
            else:
                num_str = format_auto_float(value,
                                            self._max_value_len,
                                            allow_exponent_notation=False)

            return f'{num_str.strip()}{self._unit_separator}{unit_full.strip()}'

        # no more prefixes left
        num_str = format_auto_float(value,
                                    self._max_value_len,
                                    allow_exponent_notation=False)
        unit_full = ('?' * max([len(p) for p in self._prefixes if p])) + self._unit
        return f'{num_str.strip()}{self._unit_separator}{unit_full.strip()}'


# ---------------------------------------------------------------------------
# Preset formatters
# ---------------------------------------------------------------------------

PREFIXES_SI = ['y', 'z', 'a', 'f', 'p', 'n', 'μ', 'm', None, 'k', 'M', 'G', 'T', 'P', 'E', 'Z', 'Y']
"""
Prefix presets used by default module formatters. Can be
useful if you are building your own formatter.
"""
PREFIX_ZERO_SI = 8
"""
Index of prefix which will be used as default, i.e. without 
multiplying coefficients.
"""

_formatter_si_metric = PrefixedUnitFormatter(
    max_value_len=4,
    truncate_frac=False,
    unit='',
    unit_separator=' ',
    mcoef=1000.0,
    prefixes=PREFIXES_SI,
    prefix_zero_idx=PREFIX_ZERO_SI,
)
"""
Suitable for formatting any SI unit with values
from approximately ``10^-27`` to ``10^27``.

``max_value_len`` must be at least **4**, because it's a
minimum requirement for formatting values from ``999`` to ``-999``.
Next number to 999 is ``1000``, which will be formatted as "1k".

Total maximum length is ``max_value_len + 3``, which is **7** 
(+3 is from separator, unit and prefix, assuming all of them
have 1-char width). Without unit (default) it's **6**.
"""

_formatter_si_binary = PrefixedUnitFormatter(
    max_value_len=5,
    truncate_frac=True,
    unit='b',
    unit_separator=' ',
    mcoef=1024.0,
    prefixes=PREFIXES_SI,
    prefix_zero_idx=PREFIX_ZERO_SI,
)
"""
Similar to `_formatter_si_metric`, but this formatter differs in 
one aspect.  Given a variable with default value = ``995``,
formatting it's value results in "995 b". After increasing it 
by ``20`` we'll have ``1015``, but it's still not enough to become 
a kilobyte -- so returned value will be "1015 b". Only after one 
more increase (at ``1024`` and more) the value will be in a form
of "1.00 kb".  

So, in this case ``max_value_len`` must be at least **5** (not 4),
because it's a minimum requirement for formatting values from ``1023`` 
to ``-1023``.

Total maximum length is ``max_value_len + 3`` = **8** (+3 is from separator,
unit and prefix, assuming all of them have 1-char width).
"""



# -----------------------------------------------------------------------------
# es7s/kolombos [Escape sequences and control characters visualiser]
# (C) 2022 A. Shavykin <0.delameter@gmail.com>
# -----------------------------------------------------------------------------
from . import AbstractFormatter, BinaryFormatter, TextFormatter
from ...settings import SettingsManager


class FormatterFactory:
    @staticmethod
    def create(*args) -> AbstractFormatter:
        read_mode = SettingsManager.app_settings.read_mode

        if read_mode.is_text:
            return TextFormatter(*args)
        elif read_mode.is_binary:
            return BinaryFormatter(*args)

        raise RuntimeError(f'Invalid read mode {read_mode}')



# -----------------------------------------------------------------------------
# es7s/kolombos [Escape sequences and control characters visualiser]
# (C) 2022 A. Shavykin <0.delameter@gmail.com>
# -----------------------------------------------------------------------------
from __future__ import annotations

from argparse import Namespace
from typing import Any

from .byteio import CharClass, DisplayMode, ReadMode, MarkerDetailsEnum


class Settings(Namespace):
    def __init__(self, **kwargs: Any):
        super().__init__(**kwargs)

        # auto: bool = True
        self.binary: bool = False
        self.buffer: int|None = None  # auto
        self.columns: int|None = None  # auto
        self.decimal_offsets: bool = False
        self.debug: int = 0
        self.decode: bool = False  # @TODO get rid of; will be useless after breakdown mode impl
        self.filename: str|None = None
        self.focus_control: bool = False
        self.focus_esc: bool = False
        self.focus_space: bool = False
        self.focus_utf8: bool = False
        self.focus_binary: bool = False
        self.focus_printable: bool = False
        self.ignore_control: bool = False
        self.ignore_esc: bool = False
        self.ignore_space: bool = False
        self.ignore_utf8: bool = False
        self.ignore_binary: bool = False
        self.ignore_printable: bool = False
        self.hide_ignored: bool = False  # TODO
        self.legend: bool = False
        self.marker: int = MarkerDetailsEnum.DEFAULT.value
        self.max_bytes: int|None = None  # no limit
        self.max_lines: int|None = None  # no limit
        self.color_markers: bool = False
        self.no_line_numbers: bool = False
        self.no_separators: bool = False
        self.no_offsets: bool = False
        self.squash_ignored: bool = False  # TODO
        self.text: bool = True
        self.version: bool = False

    @property
    def read_mode(self) -> ReadMode:
        if self.binary:
            return ReadMode.BINARY
        elif self.text:
            return ReadMode.TEXT
        # auto
        raise ValueError('Read mode is undefined')

    def get_char_class_display_mode(self, char_class: CharClass) -> DisplayMode:
        attr_ignored = f'ignore_{char_class.value}'
        attr_focused = f'focus_{char_class.value}'
        if not hasattr(self, attr_ignored):
            raise KeyError(f'Ignore attribute for char class {char_class} not found in settings (tried {attr_ignored})')
        if not hasattr(self, attr_focused):
            raise KeyError(f'Focus attribute for char class {char_class} not found in settings (tried {attr_focused})')

        if getattr(self, attr_ignored):
            return DisplayMode.IGNORED
        if getattr(self, attr_focused):
            return DisplayMode.FOCUSED
        return DisplayMode.DEFAULT

    @property
    def effective_marker_details(self) -> MarkerDetailsEnum:
        if self.binary:
            return MarkerDetailsEnum.BINARY_STRICT

        if self.marker <= 0:
            return MarkerDetailsEnum.NO_DETAILS
        if self.marker == 1:
            return MarkerDetailsEnum.BRIEF_DETAILS
        return MarkerDetailsEnum.FULL_DETAILS

    @property
    def effective_print_offsets(self) -> bool:
        if self.debug:
            return True
        return not self.no_offsets

    @property
    def debug_settings(self) -> int:
        return self.debug >= 3

    @property
    def debug_settings_derived(self) -> int:
        return self.debug >= 4

    @property
    def debug_buffer_contents(self) -> int:
        return self.debug >= 3

    @property
    def debug_buffer_contents_full(self) -> int:
        return self.debug >= 4


class SettingsManager:
    app_settings: Settings

    @staticmethod
    def init():
        SettingsManager.app_settings = Settings()



# -----------------------------------------------------------------------------
# es7s/kolombos [Escape sequences and control characters visualiser]
# (C) 2022 A. Shavykin <0.delameter@gmail.com>
# -----------------------------------------------------------------------------
from pytermor import SequenceSGR, Seqs, ansi

from . import Template, ControlCharGenericTemplate, Utf8SequenceTemplate, \
    PrintableCharTemplate, NewlineTemplate, EscapeSequenceTemplate, EscapeSequenceSGRTemplate
from .. import CharClass, DisplayMode, ReadMode, LabelPOV, OpeningSeqPOV


class TemplateRegistry:
    def __init__(self):
        c_cc = CharClass.CONTROL_CHAR
        self.CONTROL_CHAR = ControlCharGenericTemplate(Seqs.RED, 'Ɐ')   # 0x01-0x07, 0x0e-x1a, 0x1c-0x1f
        self.CONTROL_CHAR_NULL = Template(c_cc, Seqs.HI_RED, 'Ø')       # 0x00
        self.CONTROL_CHAR_BACKSPACE = Template(c_cc, Seqs.RED, '←')     # 0x08
        self.CONTROL_CHAR_DELETE = Template(c_cc, Seqs.RED, '→')        # 0x7f
        self.CONTROL_CHAR_ESCAPE = Template(c_cc, Seqs.HI_YELLOW, '∌')  # 0x1b

        c_ws = CharClass.WHITESPACE
        op_ws: OpeningSeqPOV = OpeningSeqPOV(Seqs.HI_CYAN, {DisplayMode.FOCUSED: Seqs.BG_CYAN + Seqs.BLACK})
        self.WHITESPACE_TAB = Template(c_ws, op_ws, LabelPOV('⇥', {ReadMode.TEXT: '⇥\t'}))        # 0x09
        self.WHITESPACE_NEWLINE = NewlineTemplate(op_ws, '↵')                                     # 0x0a
        self.WHITESPACE_VERT_TAB = Template(c_ws, op_ws, '⤓')                                     # 0x0b
        self.WHITESPACE_FORM_FEED = Template(c_ws, op_ws, '↡')                                    # 0x0c
        self.WHITESPACE_CARR_RETURN = Template(c_ws, op_ws, '⇤')                                  # 0x0d @FIXME #13
        self.WHITESPACE_SPACE = Template(c_ws, op_ws, LabelPOV('␣', {DisplayMode.FOCUSED: '·'}))  # 0x20

        self.ESCAPE_SEQ_SGR_0 = EscapeSequenceTemplate(SequenceSGR.init_color_indexed(255) +
                                                       SequenceSGR.init_color_indexed(16, True), 'θ')  # \e[m
        self.ESCAPE_SEQ_SGR = EscapeSequenceSGRTemplate(SequenceSGR.init_color_indexed(210) +
                                                        SequenceSGR.init_color_indexed(16, True), 'ǝ')  # \e[ (0x30-3f) (0x20-2f) m
        self.ESCAPE_SEQ_CSI = EscapeSequenceTemplate(Seqs.HI_GREEN, 'Ͻ')   # \e[ (0x30-3f) (0x20-2f) ...
        self.ESCAPE_SEQ_NF = EscapeSequenceTemplate(Seqs.GREEN, 'ꟻ')       # \e (0x20-2f) ...
        self.ESCAPE_SEQ_FP = EscapeSequenceTemplate(Seqs.YELLOW, 'ꟼ')      # \e (0x30-3f)
        self.ESCAPE_SEQ_FE = EscapeSequenceTemplate(Seqs.YELLOW, 'Ǝ')      # \e (0x40-5f)
        self.ESCAPE_SEQ_FS = EscapeSequenceTemplate(Seqs.YELLOW, 'Ꙅ')      # \e (0x60-7e)

        self.UTF_8_SEQ = Utf8SequenceTemplate(Seqs.HI_BLUE, LabelPOV('', {ReadMode.BINARY: '▯'}))

        c_bin = CharClass.BINARY_DATA
        self.BINARY_DATA = Template(c_bin, Seqs.MAGENTA, LabelPOV('Ḇ', {ReadMode.BINARY: '▯'}))  # 0x80-0xff

        op_pr: OpeningSeqPOV = OpeningSeqPOV(SequenceSGR.init_color_indexed(102), {DisplayMode.FOCUSED: Seqs.BG_WHITE + Seqs.BLACK})
        self.PRINTABLE_CHAR = PrintableCharTemplate(op_pr, '')  # 0x21-0x7e
